# coding=utf-8

from flask import Response

__all__ = ['RequestFilter', 'RequestFilterMetadata', 'RequestFilterChain']


class RequestFilter:
    def before_request(self):
        pass

    def after_request(self, response: Response):
        return response


class RequestFilterMetadata:
    def __init__(self, request_filter: RequestFilter):
        self.before_request = None
        self.after_request = None
        d = set(type(request_filter).__dict__.keys())
        if 'before_request' in d:
            self.before_request = request_filter.before_request
        if 'after_request' in d:
            self.after_request = request_filter.after_request


class RequestFilterChain(RequestFilter):
    def __init__(self):
        self._filters = []

    def add_request_filter(self, request_filter: RequestFilter):
        request_filter.metadata = RequestFilterMetadata(request_filter)
        self._filters.append(request_filter)

    def before_request(self):
        for f in self._filters:
            if f.metadata.before_request:
                f.metadata.before_request()

    def after_request(self, response):
        for f in self._filters:
            if f.metadata.after_request:
                response = f.metadata.after_request(response)
        return response



# coding=utf-8

__all__ = ['ASYNC_ANNOTATION_PROCESSOR',
           'SCHEDULED_ANNOTATION_PROCESSOR']

ASYNC_ANNOTATION_PROCESSOR = '__async_annotation_processor'
SCHEDULED_ANNOTATION_PROCESSOR = '__scheduled_annotation_processor'



# coding=utf-8

from abc import ABCMeta, abstractmethod

from guniflask.security.authentication import Authentication

__all__ = ['AuthenticationManager']


class AuthenticationManager(metaclass=ABCMeta):
    @abstractmethod
    def authenticate(self, authentication: Authentication) -> Authentication:
        pass



# coding=utf-8

from flask import jsonify

from guniflask.web.bind_annotation import blueprint, get_route, post_route
from guniflask.oauth2.abstract_endpoint import AbstractEndpoint
from guniflask.oauth2.token_converter import JwtAccessTokenConverter
from guniflask.oauth2.errors import InvalidClientError, InvalidRequestError, InvalidGrantError, \
    UnsupportedGrantTypeError, OAuth2AccessDeniedError, OAuth2Error
from guniflask.security.errors import InsufficientAuthenticationError, AuthenticationError
from guniflask.security.context import SecurityContext
from guniflask.security.authentication import Authentication
from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.oauth2.request_factory import DefaultOAuth2RequestValidator
from guniflask.oauth2.oauth2_utils import OAuth2Utils
from guniflask.oauth2.token import OAuth2AccessToken
from guniflask.oauth2.token_granter import TokenGranter
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.web.filter_annotation import error_handler

__all__ = ['TokenEndpoint', 'TokenKeyEndpoint']


@blueprint
class TokenEndpoint(AbstractEndpoint):

    def __init__(self, token_granter: TokenGranter, client_details_service: ClientDetailsService):
        super().__init__(token_granter, client_details_service)
        self.oauth2_request_validator = DefaultOAuth2RequestValidator()

    @post_route('/oauth/token')
    def post_access_token(self):
        auth = SecurityContext.get_authentication()
        if auth is None:
            raise InsufficientAuthenticationError('Authentication required')

        client_id = self._get_client_id(auth)
        authenticated_client = self.client_details_service.load_client_details_by_client_id(client_id)

        parameters = OAuth2Utils.get_request_parameters()
        token_request = self.oauth2_request_factory.create_token_request(parameters, authenticated_client)
        if client_id:
            if client_id != token_request.client_id:
                raise InvalidClientError('Given client ID does not match authenticated client')
        self.oauth2_request_validator.validate_scope(token_request, authenticated_client)
        if not token_request.grant_type:
            raise InvalidRequestError('Missing grant type')
        if token_request.grant_type == 'implicit':
            raise InvalidGrantError('Implicit grant type not supported from token endpoint')

        if self._is_auth_code_request(parameters):
            # The scope was requested or determined during the authorization step
            if token_request.scope:
                token_request.scope = set()
        if self._is_refresh_token_request(parameters):
            # A refresh token has its own default scopes, so we should ignore any added by the factory here
            token_request.scope = OAuth2Utils.parse_parameter_list(parameters.get(OAuth2Utils.SCOPE))

        token = self.token_granter.grant(token_request.grant_type, token_request)
        if token is None:
            raise UnsupportedGrantTypeError('Unsupported grant type: {}'.format(token_request.grant_type))
        return self._get_response(token)

    def _get_client_id(self, auth: Authentication):
        if not auth.is_authenticated:
            raise InvalidClientError('The client is not authenticated')
        client_id = auth.name
        if isinstance(auth, OAuth2Authentication):
            client_id = auth.oauth2_request.client_id
        return client_id

    def _is_auth_code_request(self, parameters: dict):
        return parameters.get('grant_type') == 'authorization_code' and parameters.get('code') is not None

    def _is_refresh_token_request(self, parameters: dict):
        return parameters.get('grant_type') == 'refresh_token' and parameters.get('refresh_token') is not None

    def _get_response(self, token: OAuth2AccessToken):
        resp = jsonify(token.to_dict())
        resp.headers['Cache-Control'] = 'no-store'
        resp.headers['Pragma'] = 'no-cache'
        return resp

    @error_handler(OAuth2Error)
    def handle_oauth2_error(self, e):
        return str(e), 400

    @error_handler(AuthenticationError)
    def handle_authentication_error(self, e):
        return str(e), 400


@blueprint
class TokenKeyEndpoint:
    def __init__(self, token_converter: JwtAccessTokenConverter):
        self.token_converter = token_converter

    @get_route('/oauth/token_key')
    def get_token_key(self):
        auth = SecurityContext.get_authentication()
        if (auth is None or not auth.is_authenticated) and not self.token_converter.is_public:
            raise OAuth2AccessDeniedError('Authentication required to see a shared key')
        return jsonify(self.token_converter.get_key())

    @error_handler(OAuth2Error)
    def handle_oauth2_error(self, e):
        return str(e), 400



# coding=utf-8

from guniflask.oauth2.token_granter import AbstractTokenGranter
from guniflask.oauth2.token_service import AuthorizationServerTokenServices
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.request_factory import OAuth2RequestFactory
from guniflask.oauth2.request import TokenRequest
from guniflask.oauth2.token import OAuth2AccessToken
from guniflask.oauth2.client_details import ClientDetails

__all__ = ['RefreshTokenGranter']


class RefreshTokenGranter(AbstractTokenGranter):
    GRANT_TYPE = 'refresh_token'

    def __init__(self, token_services: AuthorizationServerTokenServices,
                 client_details_service: ClientDetailsService,
                 request_factory: OAuth2RequestFactory):
        super().__init__(token_services, client_details_service, request_factory, self.GRANT_TYPE)

    def _get_access_token(self, token_request: TokenRequest, client: ClientDetails) -> OAuth2AccessToken:
        refresh_token_value = token_request.request_parameters.get('refresh_token')
        return self.token_services.refresh_access_token(refresh_token_value, token_request)



# coding=utf-8

from typing import Optional
from abc import ABCMeta, abstractmethod
import datetime as dt
import re

from guniflask.security.authentication import Authentication
from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.user_details_service import UserDetailsService
from guniflask.security.jwt import JwtHelper
from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.oauth2.token import OAuth2AccessToken, OAuth2RefreshToken
from guniflask.oauth2.request import OAuth2Request
from guniflask.oauth2.errors import InvalidTokenError

__all__ = ['AccessTokenConverter', 'TokenEnhancer', 'UserAuthenticationConverter',
           'DefaultAccessTokenConverter', 'JwtAccessTokenConverter']


class AccessTokenConverter(metaclass=ABCMeta):
    """
    Converter interface for token service implementations that store authentication data inside the token.
    """

    AUD = 'aud'
    CLIENT_ID = 'client_id'
    EXP = 'exp'
    JTI = 'jti'
    ATI = 'ati'
    SCOPE = OAuth2AccessToken.SCOPE
    AUTHORITIES = 'authorities'

    @abstractmethod
    def convert_access_token(self, token: OAuth2AccessToken, authentication: OAuth2Authentication) -> dict:
        pass

    @abstractmethod
    def extract_access_token(self, token_value: str, data: dict) -> OAuth2AccessToken:
        pass

    @abstractmethod
    def extract_authentication(self, data: dict) -> OAuth2Authentication:
        pass


class DefaultAccessTokenConverter(AccessTokenConverter):
    def __init__(self):
        self.user_token_converter = UserAuthenticationConverter()

    def convert_access_token(self, token: OAuth2AccessToken, authentication: OAuth2Authentication) -> dict:
        result = {}
        client_token = authentication.oauth2_request

        if not authentication.is_client_only:
            result.update(self.user_token_converter.convert_user_authentication(authentication.user_authentication))
        else:
            if client_token.authorities:
                result[self.AUTHORITIES] = client_token.authorities

        if token.scope:
            result[self.SCOPE] = token.scope
        if self.JTI in token.additional_information:
            result[self.JTI] = token.additional_information[self.JTI]
        if token.expiration:
            result[self.EXP] = int(token.expiration.timestamp())

        result.update(token.additional_information)

        result[self.CLIENT_ID] = client_token.client_id
        if client_token.resource_ids:
            result[self.AUD] = client_token.resource_ids
        return result

    def extract_access_token(self, token_value: str, data: dict) -> OAuth2AccessToken:
        token = OAuth2AccessToken(token_value)
        info = dict(data)
        if self.SCOPE in info:
            info.pop(self.SCOPE)
        if self.EXP in info:
            info.pop(self.EXP)
        if self.CLIENT_ID in info:
            info.pop(self.CLIENT_ID)
        if self.AUD in info:
            info.pop(self.AUD)
        if self.EXP in data:
            token.expiration = dt.datetime.fromtimestamp(data[self.EXP])
        if self.JTI in data:
            info[self.JTI] = data[self.JTI]
        token.scope = self._extract_scope(data)
        token.additional_information = info
        return token

    def extract_authentication(self, data: dict) -> OAuth2Authentication:
        params = {}
        scope = self._extract_scope(data)
        user = self.user_token_converter.extract_authentication(data)
        client_id = data.get(self.CLIENT_ID)
        params[self.CLIENT_ID] = client_id
        resource_ids = self._get_audience(data)
        authorities = None
        if user is None and self.AUTHORITIES in data:
            authorities = set(data[self.AUTHORITIES])
        request = OAuth2Request(request_parameters=params,
                                client_id=client_id,
                                authorities=authorities,
                                approved=True,
                                scope=scope,
                                resource_ids=resource_ids)
        return OAuth2Authentication(request, user_authentication=user)

    def _extract_scope(self, data: dict):
        scope = set()
        if self.SCOPE in data:
            s = data[self.SCOPE]
            if isinstance(s, str):
                scope.update(s.split(' '))
            elif hasattr(s, '__iter__'):
                scope.update(s)
        return scope

    def _get_audience(self, data: dict):
        result = set()
        if self.AUD not in data:
            return result
        auds = data[self.AUD]
        if isinstance(auds, str):
            result.add(auds)
        elif hasattr(auds, '__iter__'):
            result.update(auds)
        return set(auds)


class TokenEnhancer(metaclass=ABCMeta):

    @abstractmethod
    def enhance(self, access_token: OAuth2AccessToken, authentication: OAuth2Authentication) -> OAuth2AccessToken:
        pass


class UserAuthenticationConverter:
    """
    Utility interface for converting a user authentication to and from a dict.
    """

    USERNAME = 'username'
    AUTHORITIES = AccessTokenConverter.AUTHORITIES

    def __init__(self):
        self.default_authorities = set()
        self.user_details_service: UserDetailsService = None

    def convert_user_authentication(self, user_authentication: Authentication) -> dict:
        result = {}
        result[self.USERNAME] = user_authentication.name
        if user_authentication.authorities:
            result[self.AUTHORITIES] = user_authentication.authorities
        return result

    def extract_authentication(self, data: dict) -> Optional[Authentication]:
        if self.USERNAME in data:
            principal = data[self.USERNAME]
            authorities = self._get_authorities(data)
            if self.user_details_service is not None:
                user = self.user_details_service.load_user_by_username(str(principal))
                authorities = user.authorities
                principal = user
            return UserAuthentication(principal, authorities=authorities)

    def _get_authorities(self, data: dict):
        if self.AUTHORITIES not in data:
            return self.default_authorities
        authorities = set(data[self.AUTHORITIES])
        return authorities


class JwtAccessTokenConverter(AccessTokenConverter, TokenEnhancer):
    TOKEN_ID = AccessTokenConverter.JTI
    ACCESS_TOKEN_ID = AccessTokenConverter.ATI

    def __init__(self):
        self.token_converter = DefaultAccessTokenConverter()
        self.signing_algorithm = 'HS256'
        self.signing_key = JwtHelper.generate_jwt_secret()
        self.verifying_key = self.signing_key

    def convert_access_token(self, token: OAuth2AccessToken, authentication: OAuth2Authentication) -> dict:
        return self.token_converter.convert_access_token(token, authentication)

    def extract_access_token(self, token_value: str, data: dict) -> OAuth2AccessToken:
        return self.token_converter.extract_access_token(token_value, data)

    def extract_authentication(self, data: dict) -> OAuth2Authentication:
        return self.token_converter.extract_authentication(data)

    def enhance(self, access_token: OAuth2AccessToken, authentication: OAuth2Authentication) -> OAuth2AccessToken:
        result = access_token.copy()
        info = dict(access_token.additional_information)
        token_id = result.value
        if self.TOKEN_ID not in info:
            info[self.TOKEN_ID] = token_id
        else:
            token_id = str(info[self.TOKEN_ID])
        result.additional_information = info
        result.value = self.encode(result, authentication)

        refresh_token = result.refresh_token
        if refresh_token:
            encoded_refresh_token = access_token.copy()
            encoded_refresh_token.value = refresh_token.value
            encoded_refresh_token.expiration = refresh_token.expiration
            try:
                claims = JwtHelper.decode_jwt(refresh_token.value, self.verifying_key, self.signing_algorithm)
                if self.TOKEN_ID in claims:
                    encoded_refresh_token.value = str(claims[self.TOKEN_ID])
            except Exception:
                pass
            refresh_token_info = dict(access_token.additional_information)
            refresh_token_info[self.TOKEN_ID] = encoded_refresh_token.value
            refresh_token_info[self.ACCESS_TOKEN_ID] = token_id
            encoded_refresh_token.additional_information = refresh_token_info
            token = OAuth2RefreshToken(self.encode(encoded_refresh_token, authentication),
                                       expiration=refresh_token.expiration)
            result.refresh_token = token
        return result

    def encode(self, access_token: OAuth2AccessToken, authentication: OAuth2Authentication) -> str:
        payload = self.token_converter.convert_access_token(access_token, authentication)
        token = JwtHelper.encode_jwt(payload, self.signing_key, self.signing_algorithm)
        return token

    def decode(self, token_value: str) -> dict:
        try:
            payload = JwtHelper.decode_jwt(token_value, self.verifying_key, algorithm=self.signing_algorithm)
        except Exception as e:
            raise InvalidTokenError(e)
        return payload

    def is_refresh_token(self, token):
        return self.ACCESS_TOKEN_ID in token.additional_information

    @property
    def is_public(self):
        if re.match(r'^(RS|ES|PS)\d+$', self.signing_algorithm) is not None:
            return True
        return False

    def get_key(self):
        result = {'alg': self.signing_algorithm,
                  'value': self.verifying_key}
        return result



# coding=utf-8

from guniflask.security.authentication_token import AuthenticationToken

__all__ = ['PreAuthenticatedToken']


class PreAuthenticatedToken(AuthenticationToken):
    def __init__(self, principal, credentials=None, authorities=None):
        super().__init__(authorities=authorities)
        self._principal = principal
        self._credentials = credentials
        if authorities is not None:
            self.authenticate(True)

    @property
    def principal(self):
        return self._principal

    @property
    def credentials(self):
        return self._credentials



# coding=utf-8

import os
import logging
from os.path import isfile, join
import copy
from collections import MutableMapping
from importlib import import_module

from flask import current_app
from werkzeug.local import LocalProxy

log = logging.getLogger(__name__)

__all__ = ['settings', 'app_default_settings',
           'AppConfig', 'load_config', 'load_profile_config',
           'Settings']

settings = LocalProxy(lambda: current_app.extensions['settings'])

app_default_settings = {
    # Flask-SQLAlchemy
    'SQLALCHEMY_TRACK_MODIFICATIONS': False
}


class AppConfig:
    def __init__(self, app):
        s = Settings(self._load_app_settings(app.name))
        app.extensions['settings'] = s
        self.app = app

    def init_app(self):
        app_module = import_module(self.app.name + '.app')
        s = self.app_settings(self.app)

        for k, v in s.items():
            if k.isupper():
                self.app.config[k] = v
        self._set_app_default_settings(self.app)

    @property
    def settings(self):
        return self.app_settings(current_app)

    def app_settings(self, app):
        return app.extensions.get('settings')

    def _load_app_settings(self, app_name):
        c = {}
        conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
        active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES')
        kwargs = self._get_default_settings_from_env()
        if conf_dir:
            c = load_profile_config(conf_dir, app_name, profiles=active_profiles, **kwargs)
        c.update(kwargs)
        s = {}
        for name in c:
            if not name.startswith('_'):
                s[name] = c[name]
        return s

    def _get_default_settings_from_env(self):
        kwargs = {'home': os.environ.get('GUNIFLASK_HOME', os.curdir),
                  'project_name': os.environ.get('GUNIFLASK_PROJECT_NAME')}
        if os.environ.get('GUNIFLASK_DEBUG'):
            kwargs['debug'] = True
        else:
            kwargs['debug'] = False
        return kwargs

    def _set_app_default_settings(self, app):
        for k, v in app_default_settings.items():
            app.config.setdefault(k, v)


def load_config(fname, **kwargs):
    if fname is None or not isfile(fname):
        raise FileNotFoundError("Cannot find configuration file '{}'".format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    cfg.update(kwargs)
    exec(code, cfg, cfg)
    return cfg


def load_profile_config(conf_dir, name, profiles=None, **kwargs):
    pc = load_config(join(conf_dir, name + '.py'), **kwargs)
    if profiles:
        profiles = profiles.split(',')
        for profile in reversed(profiles):
            if profile:
                pc_file = join(conf_dir, name + '_' + profile + '.py')
                if isfile(pc_file):
                    c = load_config(pc_file, **kwargs)
                    pc.update(c)
        pc['active_profiles'] = list(profiles)
    return pc


class Settings(MutableMapping):
    def __init__(self, __values=None, **kwargs):
        self.attributes = {}
        self.update(__values, **kwargs)

    def __getitem__(self, name):
        if name not in self:
            return None
        return self.attributes[name]

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def get_by_prefix(self, prefix, default=None):
        s = prefix.split('.')
        obj = self
        for i in s:
            if hasattr(obj, i):
                obj = getattr(obj, i)
            elif hasattr(obj, '__getitem__') and hasattr(obj, '__contains__'):
                if i in obj:
                    obj = obj[i]
                else:
                    obj = None
            else:
                obj = None
            if obj is None:
                break
        return obj or default

    def getbool(self, name, default=None):
        v = self.get(name, default)
        return getbool(v)

    def getint(self, name, default=None):
        v = self.get(name, default)
        return getint(v)

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        return getfloat(v)

    def getlist(self, name, default=None):
        v = self.get(name, default)
        return getlist(v)

    def __setitem__(self, name, value):
        self.attributes[name] = value

    def set(self, name, value):
        self[name] = value

    def setdefault(self, k, default=None):
        if k not in self:
            self[k] = default

    def update(self, __values=None, **kwargs):
        if __values is not None:
            if isinstance(__values, Settings):
                for name in __values:
                    self[name] = __values[name]
            else:
                for name, value in __values.items():
                    self[name] = value
        for k, v in kwargs.items():
            self[k] = v

    def delete(self, name):
        del self.attributes[name]

    def __delitem__(self, name):
        del self.attributes[name]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)


def getbool(v):
    try:
        return bool(int(v))
    except (ValueError, TypeError):
        if v in ("True", "true"):
            return True
        if v in ("False", "false"):
            return False
    return None


def getint(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        pass
    return None


def getfloat(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        pass
    return None


def getlist(v):
    if v is None:
        return None
    if isinstance(v, str):
        v = v.split(",")
    elif not hasattr(v, "__iter__"):
        v = [v]
    return list(v)



# coding=utf-8

from flask import request, g

__all__ = ['OAuth2AuthenticationDetails']


class OAuth2AuthenticationDetails:
    ACCESS_TOKEN_VALUE = __qualname__ + '.access_token_value'
    ACCESS_TOKEN_TYPE = __qualname__ + '.access_token_type'

    def __init__(self):
        self.remote_address = request.remote_addr
        self.token_type = g.get(self.ACCESS_TOKEN_TYPE)
        self.token_value = g.get(self.ACCESS_TOKEN_VALUE)



# coding=utf-8

from abc import ABCMeta, abstractmethod
from typing import Collection

from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.token_service import AuthorizationServerTokenServices, DefaultTokenServices
from guniflask.oauth2.token_store import TokenStore, JwtTokenStore
from guniflask.oauth2.token_converter import TokenEnhancer, AccessTokenConverter, JwtAccessTokenConverter, \
    DefaultAccessTokenConverter
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.oauth2.request_factory import OAuth2RequestFactory, OAuth2RequestValidator, \
    DefaultOAuth2RequestValidator, DefaultOAuth2RequestFactory
from guniflask.oauth2.token_granter import TokenGranter, CompositeTokenGranter
from guniflask.oauth2.password_grant import PasswordTokenGranter
from guniflask.oauth2.client_grant import ClientCredentialsTokenGranter
from guniflask.oauth2.refresh_grant import RefreshTokenGranter
from guniflask.security_config.http_security import HttpSecurity
from guniflask.security.password_encoder import PasswordEncoder
from guniflask.oauth2.client_details_user_details_service import ClientDetailsUserDetailsService
from guniflask.security_config.authentication_manager_builder import AuthenticationManagerBuilder
from guniflask.oauth2.client_details_service import InMemoryClientDetailsService
from guniflask.security.user_details_service import UserDetailsService
from guniflask.security_config.security_configurer import SecurityConfigurerAdapter
from guniflask.oauth2_config.client_details_service_configurer import ClientDetailsServiceConfigurer

__all__ = ['AuthorizationServerConfigurer',
           'AuthorizationServerConfigurerAdapter',
           'AuthorizationServerEndpointsConfigurer',
           'AuthorizationServerSecurityConfigurer']


class AuthorizationServerEndpointsConfigurer:
    def __init__(self):
        self._client_details_service: ClientDetailsService = None
        self._token_services: AuthorizationServerTokenServices = None
        self._token_store: TokenStore = None
        self._token_enhancer: TokenEnhancer = None
        self._access_token_converter: AccessTokenConverter = None
        self._token_granter: TokenGranter = None
        self._request_factory: OAuth2RequestFactory = None
        self._request_validator: OAuth2RequestValidator = None
        self._authentication_manager: AuthenticationManager = None
        self._token_services_override = False
        self._user_details_service: UserDetailsService = None
        self._user_details_service_override = False

    @property
    def token_services(self) -> AuthorizationServerTokenServices:
        if self._token_services is None:
            self._token_services = self._create_default_token_services()
        return self._token_services

    @property
    def token_store(self) -> TokenStore:
        return self._get_token_store()

    @property
    def token_enhancer(self) -> TokenEnhancer:
        return self._token_enhancer

    @property
    def access_token_converter(self) -> AccessTokenConverter:
        return self._get_access_token_converter()

    @property
    def client_details_service(self) -> ClientDetailsService:
        return self._get_client_details_service()

    @property
    def oauth2_request_factory(self) -> OAuth2RequestFactory:
        return self._get_request_factory()

    @property
    def oauth2_request_validator(self) -> OAuth2RequestValidator:
        return self._get_request_validator()

    def with_client_details_service(self, client_details_service: ClientDetailsService):
        self._client_details_service = client_details_service
        return self

    def with_token_store(self, token_store: TokenStore) -> 'AuthorizationServerEndpointsConfigurer':
        self._token_store = token_store
        return self

    def with_token_enhancer(self, token_enhancer: TokenEnhancer) -> 'AuthorizationServerEndpointsConfigurer':
        self._token_enhancer = token_enhancer
        return self

    def with_access_token_converter(self, access_token_converter: AccessTokenConverter):
        self._access_token_converter = access_token_converter
        return self

    def with_token_services(self, token_services: AuthorizationServerTokenServices
                            ) -> 'AuthorizationServerEndpointsConfigurer':
        self._token_services = token_services
        if token_services:
            self._token_services_override = True
        return self

    def with_authentication_manager(self, authentication_manager: AuthenticationManager
                                    ) -> 'AuthorizationServerEndpointsConfigurer':
        self._authentication_manager = authentication_manager
        return self

    def with_token_granter(self, token_granter: TokenGranter) -> 'AuthorizationServerEndpointsConfigurer':
        self._token_granter = token_granter
        return self

    def with_request_factory(self, request_factory: OAuth2RequestFactory) -> 'AuthorizationServerEndpointsConfigurer':
        self._request_factory = request_factory
        return self

    def with_request_validator(self, request_validator: OAuth2RequestValidator
                               ) -> 'AuthorizationServerEndpointsConfigurer':
        self._request_validator = request_validator
        return self

    def with_user_details_service(self, user_details_service: UserDetailsService):
        if user_details_service:
            self._user_details_service = user_details_service
            self._user_details_service_override = True
        return self

    @property
    def token_granter(self) -> TokenGranter:
        return self._get_token_granter()

    def _get_token_store(self) -> TokenStore:
        if self._token_store is None:
            access_token_converter = self._get_access_token_converter()
            if isinstance(access_token_converter, JwtAccessTokenConverter):
                self._token_store = JwtTokenStore(access_token_converter)
            else:
                # FIXME: default in memory token store
                self._token_store = None
        return self._token_store

    def _get_token_enhancer(self) -> TokenEnhancer:
        if self._token_enhancer is None:
            access_token_converter = self._get_access_token_converter()
            if isinstance(access_token_converter, JwtAccessTokenConverter):
                self._token_enhancer = access_token_converter
        return self._token_enhancer

    def _get_access_token_converter(self) -> AccessTokenConverter:
        if self._access_token_converter is None:
            self._access_token_converter = DefaultAccessTokenConverter()
        return self._access_token_converter

    def _get_client_details_service(self) -> ClientDetailsService:
        if self._client_details_service is None:
            self._client_details_service = InMemoryClientDetailsService()
        return self._client_details_service

    def _get_token_granter(self) -> TokenGranter:
        if self._token_granter is None:
            self._token_granter = CompositeTokenGranter(self._get_default_token_granters())
        return self._token_granter

    def _get_request_factory(self) -> OAuth2RequestFactory:
        if self._request_factory is None:
            self._request_factory = DefaultOAuth2RequestFactory(self._client_details_service)
        return self._request_factory

    def _get_request_validator(self) -> OAuth2RequestValidator:
        if self._request_validator is None:
            self._request_validator = DefaultOAuth2RequestValidator()
        return self._request_validator

    def _create_default_token_services(self) -> AuthorizationServerTokenServices:
        token_services = DefaultTokenServices(self._get_token_store())
        token_services.support_refresh_token = True
        token_services.token_enhancer = self._get_token_enhancer()
        token_services.client_details_service = self._client_details_service
        return token_services

    def _get_default_token_granters(self) -> Collection[TokenGranter]:
        token_granters = []
        token_services = self.token_services
        request_factory = self._get_request_factory()
        token_granters.append(RefreshTokenGranter(token_services, self._client_details_service, request_factory))
        token_granters.append(ClientCredentialsTokenGranter(token_services,
                                                            self._client_details_service,
                                                            request_factory))
        if self._authentication_manager is not None:
            token_granters.append(PasswordTokenGranter(self._authentication_manager,
                                                       token_services,
                                                       self._client_details_service,
                                                       request_factory))
        return token_granters


class AuthorizationServerSecurityConfigurer(SecurityConfigurerAdapter):
    def __init__(self):
        super().__init__()
        self._password_encoder: PasswordEncoder = None

    def init(self, http: HttpSecurity):
        if self._password_encoder:
            password_encoder = self._get_password_encoder()
            client_details_user_details_service = ClientDetailsUserDetailsService(self._get_client_details_service())
            client_details_user_details_service.set_password_encoder(password_encoder)
            builder: AuthenticationManagerBuilder = http.get_shared_object(AuthenticationManagerBuilder)
            builder \
                .with_user_details_service(client_details_user_details_service) \
                .with_password_encoder(password_encoder)
        else:
            http.with_user_details_service(ClientDetailsUserDetailsService(self._get_client_details_service()))

    def configure(self, http: HttpSecurity):
        pass

    def with_password_encoder(self, password_encoder: PasswordEncoder):
        self._password_encoder = password_encoder
        return self

    def _get_client_details_service(self) -> ClientDetailsService:
        return self.builder.get_shared_object(ClientDetailsService)

    def _get_password_encoder(self) -> PasswordEncoder:
        return NullablePasswordEncoder(self._password_encoder)


class NullablePasswordEncoder(PasswordEncoder):
    def __init__(self, password_encoder: PasswordEncoder):
        self.password_encoder = password_encoder

    def encode(self, raw_password: bytes) -> str:
        return self.password_encoder.encode(raw_password)

    def matches(self, raw_password: bytes, encoded_password: str) -> bool:
        if not encoded_password:
            return True
        return self.password_encoder.matches(raw_password, encoded_password)


class AuthorizationServerConfigurer(metaclass=ABCMeta):
    @abstractmethod
    def configure_endpoints(self, endpoints: AuthorizationServerEndpointsConfigurer):
        pass

    @abstractmethod
    def configure_client_details_service(self, clients: ClientDetailsServiceConfigurer):
        pass

    @abstractmethod
    def configure_security(self, security: AuthorizationServerSecurityConfigurer):
        pass


class AuthorizationServerConfigurerAdapter(AuthorizationServerConfigurer):
    def configure_endpoints(self, endpoints: AuthorizationServerEndpointsConfigurer):
        pass

    def configure_client_details_service(self, clients: ClientDetailsServiceConfigurer):
        pass

    def configure_security(self, security: AuthorizationServerSecurityConfigurer):
        pass



# coding=utf-8

import datetime as dt

import sqlalchemy

from guniflask.utils.factory import string_to_datetime

__all__ = ['model_to_dict', 'result_to_dict', 'dict_to_model', 'update_model_by_dict']


def model_to_dict(model, ignore=None, only=None, only_not_none=False):
    col_attrs = sqlalchemy.inspect(model).mapper.column_attrs
    d = {}
    tz_info = dt.datetime.now(tz=dt.timezone.utc).astimezone().tzinfo
    ignore_set = _get_field_set(ignore)
    only_set = _get_field_set(only)
    for c in col_attrs:
        if c.key not in ignore_set:
            if only and c.key not in only_set:
                continue
            v = getattr(model, c.key)
            if isinstance(v, dt.datetime) and v.tzinfo is None:
                v = v.replace(tzinfo=tz_info)
            if only_not_none and v is None:
                continue
            d[c.key] = v
    return d


def result_to_dict(result, model_cls, ignore=None, only=None, only_not_none=False):
    res = {}
    col_attrs = sqlalchemy.inspect(model_cls).mapper.column_attrs
    tz_info = dt.datetime.now(tz=dt.timezone.utc).astimezone().tzinfo
    ignore_set = _get_field_set(ignore)
    only_set = _get_field_set(only)
    for c in col_attrs:
        if c.key not in ignore_set:
            if only and c.key not in only_set:
                continue
            if hasattr(result, c.key):
                v = getattr(result, c.key)
                if isinstance(v, dt.datetime) and v.tzinfo is None:
                    v = v.replace(tzinfo=tz_info)
                if only_not_none and v is None:
                    continue
                res[c.key] = v
    return res


def dict_to_model(dict_obj, model_cls, ignore=None, only=None, only_not_none=False):
    mapper = sqlalchemy.inspect(model_cls).mapper
    columns = mapper.columns
    col_attrs = mapper.column_attrs
    ignore_set = _get_field_set(ignore)
    only_set = _get_field_set(only)
    kwargs = {}
    for c in col_attrs:
        if c.key in dict_obj and c.key not in ignore_set:
            if only and c.key not in only_set:
                continue
            v = dict_obj[c.key]
            if isinstance(columns[c.key].type, sqlalchemy.DateTime):
                v = string_to_datetime(v)
            if only_not_none and v is None:
                continue
            kwargs[c.key] = v
    model = model_cls(**kwargs)
    return model


def update_model_by_dict(model, dict_obj, ignore=None, only=None, only_not_none=False):
    mapper = sqlalchemy.inspect(model).mapper
    columns = mapper.columns
    col_attrs = mapper.column_attrs
    ignore_set = _get_field_set(ignore)
    only_set = _get_field_set(only)
    for c in col_attrs:
        if c.key in dict_obj and c.key not in ignore_set:
            if only and c.key not in only_set:
                continue
            v = dict_obj[c.key]
            if isinstance(columns[c.key].type, sqlalchemy.DateTime):
                v = string_to_datetime(v)
            if only_not_none and v is None:
                continue
            setattr(model, c.key, v)


def _get_field_set(field):
    if isinstance(field, str):
        field = [i.strip() for i in field.split(',')]
    return set(field or [])



# coding=utf-8

from abc import ABCMeta, abstractmethod

__all__ = ['InitializingBean', 'SmartInitializingSingleton', 'DisposableBean']


class InitializingBean(metaclass=ABCMeta):
    @abstractmethod
    def after_properties_set(self):
        pass


class SmartInitializingSingleton(metaclass=ABCMeta):
    @abstractmethod
    def after_singletons_instantiated(self):
        pass


class DisposableBean(metaclass=ABCMeta):
    @abstractmethod
    def destroy(self):
        pass



# coding=utf-8

from abc import ABCMeta, abstractmethod

from guniflask.security_config.security_builder import SecurityBuilder

__all__ = ['SecurityConfigurer', 'SecurityConfigurerAdapter']


class SecurityConfigurer(metaclass=ABCMeta):

    @abstractmethod
    def init(self, builder: SecurityBuilder):
        pass

    @abstractmethod
    def configure(self, builder: SecurityBuilder):
        pass


class SecurityConfigurerAdapter(SecurityConfigurer, metaclass=ABCMeta):

    def __init__(self):
        self._security_builder: SecurityBuilder = None

    def init(self, builder: SecurityBuilder):
        pass

    def configure(self, builder: SecurityBuilder):
        pass

    @property
    def builder(self):
        return self._security_builder

    @builder.setter
    def builder(self, value: SecurityBuilder):
        self._security_builder = value



# coding=utf-8

from guniflask.context.annotation import configuration, bean
from guniflask.oauth2.client_details import ClientDetails
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2_config.client_details_service_configurer import ClientDetailsServiceConfigurer

__all__ = ['ClientDetailsServiceConfiguration']


@configuration
class ClientDetailsServiceConfiguration:
    def __init__(self):
        self._configurer = ClientDetailsServiceConfigurer()

    @bean
    def client_details_configurer(self) -> ClientDetailsServiceConfigurer:
        return self._configurer

    @bean
    def client_details_service(self) -> ClientDetailsService:
        return ClientDetailsServiceDelegator(self._configurer)


class ClientDetailsServiceDelegator(ClientDetailsService):
    def __init__(self, delegate_builder: ClientDetailsServiceConfigurer):
        self._delegate_builder = delegate_builder
        self._delegate: ClientDetailsService = None

    def load_client_details_by_client_id(self, client_id) -> ClientDetails:
        if self._delegate is None:
            self._delegate = self._delegate_builder.client_details_service
            self._delegate_builder = None
        return self._delegate.load_client_details_by_client_id(client_id)



# coding=utf-8

from guniflask.security.authentication import Authentication
from guniflask.security.authentication_token import AuthenticationToken
from guniflask.oauth2.request import OAuth2Request

__all__ = ['OAuth2Authentication']


class OAuth2Authentication(AuthenticationToken):
    def __init__(self, oauth2_request: OAuth2Request, user_authentication: Authentication = None):
        authorities = None
        if user_authentication is not None:
            authorities = user_authentication.authorities
        super().__init__(authorities=authorities)
        self._oauth2_request = oauth2_request
        self._user_authentication = user_authentication

    @property
    def principal(self):
        if self._user_authentication is None:
            return self._oauth2_request.client_id
        return self.user_authentication.principal

    @property
    def credentials(self):
        return None

    @property
    def oauth2_request(self):
        return self._oauth2_request

    @property
    def user_authentication(self):
        return self._user_authentication

    @property
    def is_client_only(self):
        return self._user_authentication is None



# coding=utf-8

from .annotation import *
from .autowired_post_processor import *
from .constructor_resolver import *
from .default_factory import *
from .definition import *
from .definition_registry import *
from .errors import *
from .factory import *
from .factory_hook import *
from .factory_post_processor import *
from .name_generator import *
from .post_processor import *
from .singleton_registry import *

__all__ = (annotation.__all__ +
           autowired_post_processor.__all__ +
           constructor_resolver.__all__ +
           default_factory.__all__ +
           definition.__all__ +
           definition_registry.__all__ +
           errors.__all__ +
           factory.__all__ +
           factory_hook.__all__ +
           factory_post_processor.__all__ +
           name_generator.__all__ +
           post_processor.__all__ +
           singleton_registry.__all__)



# coding=utf-8

from guniflask.beans.factory_hook import InitializingBean
from guniflask.oauth2.token_granter import TokenGranter
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.request_factory import OAuth2RequestFactory, DefaultOAuth2RequestFactory

__all__ = ['AbstractEndpoint']


class AbstractEndpoint(InitializingBean):
    def __init__(self, token_granter: TokenGranter, client_details_service: ClientDetailsService):
        self.token_granter = token_granter
        self.client_details_service = client_details_service
        self.oauth2_request_factory: OAuth2RequestFactory = None

    def after_properties_set(self):
        assert self.token_granter is not None, 'Token granter is required'
        assert self.client_details_service is not None, 'Client details service is required'
        if self.oauth2_request_factory is None:
            self.oauth2_request_factory = DefaultOAuth2RequestFactory(self.client_details_service)



# coding=utf-8

from abc import ABCMeta, abstractmethod

from guniflask.oauth2.client_details import ClientDetails
from guniflask.oauth2.errors import NoSuchClientError

__all__ = ['ClientDetailsService', 'InMemoryClientDetailsService']


class ClientDetailsService(metaclass=ABCMeta):
    """
    A service that provides the details about an OAuth2 client.
    """

    @abstractmethod
    def load_client_details_by_client_id(self, client_id) -> ClientDetails:
        pass


class InMemoryClientDetailsService(ClientDetailsService):
    def __init__(self, client_details_store: dict = None):
        self.client_details_store = {}
        if client_details_store:
            self.client_details_store.update(client_details_store)

    def load_client_details_by_client_id(self, client_id: str) -> ClientDetails:
        if client_id not in self.client_details_store:
            raise NoSuchClientError('No client found with id = {}'.format(client_id))
        return self.client_details_store[client_id]



# coding=utf-8

from guniflask.annotation.core import Annotation, AnnotationUtils

__all__ = ['Autowired', 'autowired']


class Autowired(Annotation):
    def __init__(self):
        super().__init__()


def autowired(func):
    AnnotationUtils.add_annotation(func, Autowired())
    return func



# coding=utf-8

from .bind_annotation import *
from .blueprint_post_processor import *
from .config_constants import *
from .context import *
from .cors import *
from .filter_annotation import *
from .global_singleton_annotation import *
from .param_annotation import *
from .request_filter import *
from .scheduling_config import *

__all__ = (bind_annotation.__all__ +
           blueprint_post_processor.__all__ +
           config_constants.__all__ +
           context.__all__ +
           cors.__all__ +
           filter_annotation.__all__ +
           global_singleton_annotation.__all__ +
           param_annotation.__all__ +
           request_filter.__all__ +
           scheduling_config.__all__)



# coding=utf-8

from abc import ABCMeta, abstractmethod

__all__ = ['UserDetails']


class UserDetails(metaclass=ABCMeta):

    @property
    @abstractmethod
    def username(self):
        pass

    @property
    @abstractmethod
    def password(self):
        pass

    @property
    @abstractmethod
    def authorities(self):
        pass



# coding=utf-8

from abc import ABCMeta, abstractmethod
import datetime as dt

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from guniflask.beans.factory_hook import SmartInitializingSingleton, DisposableBean

__all__ = ['TaskScheduler', 'DefaultTaskScheduler']


class TaskScheduler(metaclass=ABCMeta):
    @abstractmethod
    def schedule(self, task, start_time: dt.datetime = None):
        pass

    @abstractmethod
    def schedule_with_cron(self, task, cron: str, start_time: dt.datetime = None):
        pass

    @abstractmethod
    def schedule_with_fixed_interval(self, task, interval: int, start_time: dt.datetime = None):
        pass


class DefaultTaskScheduler(TaskScheduler, SmartInitializingSingleton, DisposableBean):
    def __init__(self, scheduler=None):
        self._scheduler = scheduler or BackgroundScheduler()

    @property
    def scheduler(self):
        return self._scheduler

    def schedule(self, task, start_time: dt.datetime = None):
        trigger = None
        if start_time is not None:
            trigger = DateTrigger(start_time)
        self._scheduler.add_job(task, trigger=trigger)

    def schedule_with_cron(self, task, cron: str, start_time: dt.datetime = None):
        values = cron.split()
        if len(values) != 5:
            raise ValueError('Wrong number of fields: got {}, expected 5'.format(len(values)))
        cron_trigger = CronTrigger(minute=values[0], hour=values[1], day=values[2], month=values[3],
                                   day_of_week=values[4], start_date=start_time)
        self._scheduler.add_job(task, trigger=cron_trigger)

    def schedule_with_fixed_interval(self, task, interval: int, start_time: dt.datetime = None):
        interval_trigger = IntervalTrigger(seconds=interval, start_date=start_time)
        self._scheduler.add_job(task, trigger=interval_trigger)

    def start(self):
        self._scheduler.start(paused=False)

    def shutdown(self, wait=True):
        self._scheduler.shutdown(wait=wait)

    def after_singletons_instantiated(self):
        self.start()

    def destroy(self):
        self.shutdown(wait=False)



# coding=utf-8

from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.oauth2.token_service import ResourceServerTokenServices
from guniflask.beans.factory_hook import InitializingBean
from guniflask.oauth2.errors import InvalidTokenError, OAuth2AccessDeniedError, ClientRegistrationError
from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.oauth2.client_details_service import ClientDetailsService

__all__ = ['OAuth2AuthenticationManager']


class OAuth2AuthenticationManager(AuthenticationManager, InitializingBean):
    def __init__(self, ):
        self.token_services: ResourceServerTokenServices = None
        self.resource_id: str = None
        self.client_details_service: ClientDetailsService = None

    def after_properties_set(self):
        assert self.token_services is not None, 'Token services are required'

    def authenticate(self, authentication):
        if authentication is None:
            raise InvalidTokenError('Token not found')
        token_value = str(authentication.principal)
        auth = self.token_services.load_authentication(token_value)
        if auth is None:
            raise InvalidTokenError('Invalid token: {}'.format(token_value))

        resource_ids = auth.oauth2_request.resource_ids
        if self.resource_id and resource_ids and self.resource_id not in resource_ids:
            raise OAuth2AccessDeniedError(
                'Invalid token does not contain resource id ({})'.format(self.resource_id))
        self._check_client_details(auth)

        auth.details = authentication.details
        auth.authenticate(True)
        return auth

    def _check_client_details(self, auth: OAuth2Authentication):
        if self.client_details_service is not None:
            client_id = auth.oauth2_request.client_id
            try:
                client = self.client_details_service.load_client_details_by_client_id(client_id)
            except ClientRegistrationError:
                raise OAuth2AccessDeniedError('Invalid token contains invalid client id ({})'.format(client_id))
            allowed = client.scope
            for scope in auth.oauth2_request.scope:
                if scope not in allowed:
                    raise OAuth2AccessDeniedError(
                        'Invalid token contains disallowed scope ({}) for this client'.format(scope))



# coding=utf-8

__all__ = ['OAuth2Error',
           'ClientAuthenticationError',
           'InvalidRequestError',
           'InvalidClientError',
           'InvalidTokenError',
           'InvalidGrantError',
           'InvalidScopeError',
           'OAuth2AccessDeniedError',
           'UnsupportedGrantTypeError',
           'ClientRegistrationError',
           'NoSuchClientError',
           'ClientAlreadyExistsError']


class OAuth2Error(Exception):
    """
    OAuth2 exception
    """


class ClientAuthenticationError(OAuth2Error):
    """
    Client authentication exception
    """


class InvalidRequestError(ClientAuthenticationError):
    """
    Invalid request exception
    """


class InvalidClientError(ClientAuthenticationError):
    """
    Invalid client exception
    """


class InvalidTokenError(ClientAuthenticationError):
    """
    Invalid token exception
    """


class InvalidGrantError(ClientAuthenticationError):
    """
    Invalid grant exception
    """


class InvalidScopeError(OAuth2Error):
    """
    Invalid scope exception
    """


class OAuth2AccessDeniedError(OAuth2Error):
    """
    OAuth2 access denied
    """


class UnsupportedGrantTypeError(OAuth2Error):
    """
    Unsupported grant type exception
    """


class ClientRegistrationError(Exception):
    """
    Client registration exception
    """


class NoSuchClientError(ClientRegistrationError):
    """
    No such client exception
    """


class ClientAlreadyExistsError(ClientRegistrationError):
    """
    Client already exists exception
    """



# coding=utf-8

from typing import List

from guniflask.context.annotation import configuration
from guniflask.oauth2.token_service import ResourceServerTokenServices
from guniflask.oauth2.token_store import TokenStore
from guniflask.oauth2_config.authorization_server_config import AuthorizationServerEndpointsConfiguration
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2_config.resource_server_configurer import ResourceServerConfigurer, \
    ResourceServerSecurityConfigurer
from guniflask.security_config.http_security import HttpSecurity
from guniflask.security_config.web_security_configurer import WebSecurityConfigurer

__all__ = ['ResourceServerConfiguration']


@configuration
class ResourceServerConfiguration(WebSecurityConfigurer):
    def __init__(self, configurers: List[ResourceServerConfigurer] = None,
                 token_store: TokenStore = None,
                 token_services: ResourceServerTokenServices = None,
                 endpoints: AuthorizationServerEndpointsConfiguration = None):
        super().__init__()
        self._configurers = configurers
        self._token_store = token_store
        self._token_services = token_services
        self._endpoints = endpoints

    def configure_http(self, http: HttpSecurity):
        resources = ResourceServerSecurityConfigurer()
        if self._token_services:
            resources.with_token_services(self._token_services)
        else:
            if self._token_store:
                resources.with_token_store(self._token_store)
            elif self._endpoints:
                resources.with_token_store(self._endpoints.endpoints_configurer.token_store)

        if self._configurers:
            for c in self._configurers:
                c.configure_security(resources)

        http.apply(resources)

        if self._configurers:
            for c in self._configurers:
                c.configure_http(http)



# coding=utf-8

from guniflask.annotation.core import AnnotationUtils
from guniflask.context.annotation import Include
from guniflask.oauth2_config.authorization_server_config import AuthorizationServerEndpointsConfiguration, \
    AuthorizationServerSecurityConfiguration
from guniflask.oauth2_config.resource_server_config import ResourceServerConfiguration

__all__ = ['enable_authorization_server',
           'enable_resource_server']


def enable_authorization_server(func):
    values = [AuthorizationServerEndpointsConfiguration,
              AuthorizationServerSecurityConfiguration]
    AnnotationUtils.add_annotation(func, Include(values))
    return func


def enable_resource_server(func):
    values = [ResourceServerConfiguration]
    AnnotationUtils.add_annotation(func, Include(values))
    return func



# coding=utf-8

__all__ = ['BeansError',
           'NoSuchBeanDefinitionError', 'NoUniqueBeanDefinitionError',
           'BeanDefinitionStoreError',
           'BeanCreationError', 'BeanCurrentlyInCreationError',
           'BeanNotOfRequiredTypeError',
           'BeanTypeNotDeclaredError', 'BeanTypeNotAllowedError']


class BeansError(Exception):
    pass


class NoSuchBeanDefinitionError(BeansError):
    def __init__(self, bean_name):
        self.bean_name = bean_name
        super().__init__('No bean named "{}" available'.format(bean_name))


class NoUniqueBeanDefinitionError(BeansError):
    def __init__(self, bean_type):
        self.bean_type = bean_type
        super().__init__('"{}" is not a unique bean type'.format(bean_type))


class BeanDefinitionStoreError(BeansError):
    pass


class BeanCreationError(BeansError):
    def __init__(self, bean_name, message=None):
        self.bean_name = bean_name
        super().__init__(message or 'Error creating bean with name "{}"'.format(bean_name))


class BeanCurrentlyInCreationError(BeanCreationError):
    def __init__(self, bean_name):
        super().__init__(bean_name, message='Bean named "{}" is currently in creation: '
                                            'Is there an unresolvable circular reference?"'.format(bean_name))


class BeanNotOfRequiredTypeError(BeansError):
    def __init__(self, bean_name, required_type, actual_type):
        self.bean_name = bean_name
        self.required_type = required_type
        self.actual_type = actual_type
        super().__init__('Bean named "{}" is expected to be of type "{}"'
                         ' but was actually of type {}'.format(bean_name, required_type, actual_type))


class BeanTypeNotDeclaredError(BeansError):
    def __init__(self, bean_name):
        self.bean_name = bean_name
        super().__init__('The type of bean named "{}" is not declared'.format(bean_name))


class BeanTypeNotAllowedError(BeansError):
    def __init__(self, bean_name, bean_type):
        self.bean_name = bean_name
        self.bean_type = bean_type
        super().__init__('The type "{}" of bean named "{}" is not a allowed'.format(bean_type, bean_name))



# coding=utf-8

from .access_annotation import *
from .authentication import *
from .authentication_manager import *
from .authentication_provider import *
from .authentication_token import *
from .authentication_user_details_service import *
from .basic_authentication_filter import *
from .context import *
from .dao_authentication_provider import *
from .errors import *
from .jwt import *
from .jwt_provider import *
from .password_encoder import *
from .preauth_provider import *
from .preauth_token import *
from .provider_manager import *
from .user import *
from .user_context import *
from .user_details import *
from .user_details_authentication_provider import *
from .user_details_by_name_service import *
from .user_details_service import *
from .web_authentication_details import *

__all__ = (access_annotation.__all__ +
           authentication.__all__ +
           authentication_manager.__all__ +
           authentication_provider.__all__ +
           authentication_token.__all__ +
           authentication_user_details_service.__all__ +
           basic_authentication_filter.__all__ +
           context.__all__ +
           dao_authentication_provider.__all__ +
           errors.__all__ +
           jwt.__all__ +
           jwt_provider.__all__ +
           password_encoder.__all__ +
           preauth_provider.__all__ +
           preauth_token.__all__ +
           provider_manager.__all__ +
           user.__all__ +
           user_context.__all__ +
           user_details.__all__ +
           user_details_authentication_provider.__all__ +
           user_details_by_name_service.__all__ +
           user_details_service.__all__ +
           web_authentication_details.__all__)



# coding=utf-8

from guniflask.security.user_details import UserDetails
from guniflask.security.user_details_service import UserDetailsService
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.security.password_encoder import PasswordEncoder
from guniflask.oauth2.errors import NoSuchClientError
from guniflask.security.errors import UsernameNotFoundError
from guniflask.security.user import User

__all__ = ['ClientDetailsUserDetailsService']


class ClientDetailsUserDetailsService(UserDetailsService):

    def __init__(self, client_details_service: ClientDetailsService):
        self.client_details_service = client_details_service
        self.empty_password = ""

    def set_password_encoder(self, password_encoder: PasswordEncoder):
        self.empty_password = password_encoder.encode(b'')

    def load_user_by_username(self, username: str) -> UserDetails:
        try:
            client_details = self.client_details_service.load_client_details_by_client_id(username)
        except NoSuchClientError as e:
            raise UsernameNotFoundError(e)
        client_secret = client_details.client_secret
        if not client_secret:
            client_secret = self.empty_password
        return User(username=username, password=client_secret, authorities=client_details.authorities)



# coding=utf-8

from typing import Union, List

import flask_sqlalchemy

from guniflask.orm.model_utils import dict_to_model, model_to_dict, update_model_by_dict, result_to_dict

__all__ = ['FeignQuery', 'BaseModelMixin']


class FeignQuery(flask_sqlalchemy.BaseQuery):

    def with_entities(self, *entities) -> 'FeignQuery':
        return super().with_entities(*entities)

    def params(self, *args, **kwargs) -> 'FeignQuery':
        return super().params(*args, **kwargs)

    def filter(self, *criterion) -> 'FeignQuery':
        return super().filter(*criterion)

    def filter_by(self, **kwargs) -> 'FeignQuery':
        return super().filter_by(**kwargs)

    def order_by(self, *criterion) -> 'FeignQuery':
        return super().order_by(*criterion)

    def group_by(self, *criterion) -> 'FeignQuery':
        return super().group_by(*criterion)

    def having(self, criterion) -> 'FeignQuery':
        return super().having(criterion)

    def union(self, *q) -> 'FeignQuery':
        return super().union(*q)

    def union_all(self, *q) -> 'FeignQuery':
        return super().union_all(*q)

    def intersect(self, *q) -> 'FeignQuery':
        return super().intersect(*q)

    def intersect_all(self, *q) -> 'FeignQuery':
        return super().intersect_all(*q)

    def except_(self, *q) -> 'FeignQuery':
        return super().except_(*q)

    def except_all(self, *q) -> 'FeignQuery':
        return super().except_all(*q)

    def join(self, *props, **kwargs) -> 'FeignQuery':
        return super().join(*props, **kwargs)

    def outerjoin(self, *props, **kwargs) -> 'FeignQuery':
        return super().outerjoin(*props, **kwargs)

    def reset_joinpoint(self) -> 'FeignQuery':
        return super().reset_joinpoint()

    def select_from(self, *from_obj) -> 'FeignQuery':
        return super().select_from(*from_obj)

    def select_entity_from(self, from_obj) -> 'FeignQuery':
        return super().select_entity_from(from_obj)

    def slice(self, start, stop) -> 'FeignQuery':
        return super().slice(start, stop)

    def limit(self, limit) -> 'FeignQuery':
        return super().limit(limit)

    def offset(self, offset) -> 'FeignQuery':
        return super().offset(offset)

    def distinct(self, *expr) -> 'FeignQuery':
        return super().distinct(*expr)

    def prefix_with(self, *prefixes) -> 'FeignQuery':
        return super().prefix_with(*prefixes)

    def suffix_with(self, *suffixes) -> 'FeignQuery':
        return super().suffix_with(*suffixes)


class BaseModelMixin:
    query: FeignQuery

    @classmethod
    def from_dict(cls, dict_obj: dict,
                  ignore: Union[str, List] = None,
                  only: Union[str, List] = None,
                  only_not_none: bool = False):
        return dict_to_model(dict_obj, cls, ignore=ignore, only=only, only_not_none=only_not_none)

    def to_dict(self, ignore: Union[str, List] = None,
                only: Union[str, List] = None,
                only_not_none: bool = False):
        return model_to_dict(self, ignore=ignore, only=only, only_not_none=only_not_none)

    def update_by_dict(self, dict_obj: dict,
                       ignore: Union[str, List] = None,
                       only: Union[str, List] = None,
                       only_not_none: bool = False):
        return update_model_by_dict(self, dict_obj, ignore=ignore, only=only, only_not_none=only_not_none)

    @classmethod
    def result_to_dict(cls, result,
                       ignore: Union[str, List] = None,
                       only: Union[str, List] = None,
                       only_not_none: bool = False):
        return result_to_dict(result, cls, ignore=ignore, only=only, only_not_none=only_not_none)



# coding=utf-8

from .annotation import *
from .authorization_server_config import *
from .authorization_server_configurer import *
from .client_details_service_config import *
from .client_details_service_configurer import *
from .resource_server_config import *
from .resource_server_configurer import *

__all__ = (annotation.__all__ +
           authorization_server_config.__all__ +
           authorization_server_configurer.__all__ +
           client_details_service_config.__all__ +
           client_details_service_configurer.__all__ +
           resource_server_config.__all__ +
           resource_server_configurer.__all__)



# coding=utf-8

from guniflask.oauth2.client_details_service import ClientDetailsService

__all__ = ['ClientDetailsServiceConfigurer']


class ClientDetailsServiceConfigurer:
    def __init__(self):
        self._client_details_service: ClientDetailsService = None

    @property
    def client_details_service(self) -> ClientDetailsService:
        return self._client_details_service

    def with_client_details_service(self, client_details_service: ClientDetailsService
                                    ) -> 'ClientDetailsServiceConfigurer':
        self._client_details_service = client_details_service
        return self



# coding=utf-8

from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.oauth2.token_granter import AbstractTokenGranter
from guniflask.oauth2.token_service import AuthorizationServerTokenServices
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.request_factory import OAuth2RequestFactory
from guniflask.oauth2.request import TokenRequest
from guniflask.oauth2.client_details import ClientDetails
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.authentication_token import UserAuthentication
from guniflask.oauth2.errors import InvalidGrantError
from guniflask.security.errors import BadCredentialsError

__all__ = ['PasswordTokenGranter']


class PasswordTokenGranter(AbstractTokenGranter):
    GRANT_TYPE = 'password'

    def __init__(self, authentication_manager: AuthenticationManager,
                 token_services: AuthorizationServerTokenServices,
                 client_details_service: ClientDetailsService,
                 request_factory: OAuth2RequestFactory):
        super().__init__(token_services, client_details_service, request_factory, self.GRANT_TYPE)
        self.authentication_manager = authentication_manager

    def _get_oauth2_authentication(self, token_request: TokenRequest, client: ClientDetails) -> OAuth2Authentication:
        parameters = token_request.request_parameters
        username = parameters.get('username')
        password = parameters.get('password')
        parameters.pop('password')

        user_auth = UserAuthentication(username, credentials=password)
        user_auth.details = parameters
        try:
            user_auth = self.authentication_manager.authenticate(user_auth)
        except BadCredentialsError as e:
            raise InvalidGrantError(e)
        if user_auth is None or not user_auth.is_authenticated:
            raise InvalidGrantError('Could not authenticate user: {}'.format(username))
        oauth2_request = self.request_factory.create_oauth2_request_from_token_request(token_request, client)
        return OAuth2Authentication(oauth2_request, user_auth)



# coding=utf-8

from abc import ABCMeta, abstractmethod
from typing import Optional, Collection

from guniflask.oauth2.token_service import AuthorizationServerTokenServices
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.request_factory import OAuth2RequestFactory
from guniflask.oauth2.request import TokenRequest
from guniflask.oauth2.token import OAuth2AccessToken
from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.oauth2.client_details import ClientDetails
from guniflask.oauth2.errors import InvalidClientError

__all__ = ['TokenGranter', 'AbstractTokenGranter', 'CompositeTokenGranter']


class TokenGranter(metaclass=ABCMeta):

    @abstractmethod
    def grant(self, grant_type: str, token_request: TokenRequest) -> Optional[OAuth2AccessToken]:
        pass


class AbstractTokenGranter(TokenGranter):
    def __init__(self, token_services: AuthorizationServerTokenServices, client_details_service: ClientDetailsService,
                 request_factory: OAuth2RequestFactory, grant_type: str):
        self.token_services = token_services
        self.client_details_service = client_details_service
        self.request_factory = request_factory
        self.grant_type = grant_type

    def grant(self, grant_type, token_request):
        if grant_type != self.grant_type:
            return
        client_id = token_request.client_id
        client = self.client_details_service.load_client_details_by_client_id(client_id)
        self._validate_grant_type(grant_type, client)
        return self._get_access_token(token_request, client)

    def _get_access_token(self, token_request: TokenRequest, client: ClientDetails) -> OAuth2AccessToken:
        return self.token_services.create_access_token(self._get_oauth2_authentication(token_request, client))

    def _get_oauth2_authentication(self, token_request: TokenRequest,
                                   client: ClientDetails) -> OAuth2Authentication:
        oauth2_request = self.request_factory.create_oauth2_request_from_token_request(token_request, client)
        return OAuth2Authentication(oauth2_request)

    def _validate_grant_type(self, grant_type: str, client_details: ClientDetails):
        authorized_grant_types = client_details.grant_types
        if authorized_grant_types and grant_type not in authorized_grant_types:
            raise InvalidClientError('Unauthorized grant type: {}'.format(grant_type))


class CompositeTokenGranter(TokenGranter):
    def __init__(self, token_granters: Collection[TokenGranter] = None):
        self.token_granters = [] if token_granters is None else list(token_granters)

    def grant(self, grant_type: str, token_request: TokenRequest) -> Optional[OAuth2AccessToken]:
        for granter in self.token_granters:
            result = granter.grant(grant_type, token_request)
            if result is not None:
                return result

    def add_token_granter(self, token_granter):
        self.token_granters.append(token_granter)



# coding=utf-8

from typing import List

import requests

from guniflask.consul.errors import ConsulClientError

__all__ = ['Consul']


class Consul:
    api_version = 'v1'

    def __init__(self, host: str = '127.0.0.1', port: int = 8500, scheme: str = 'http'):
        self.host = host
        self.port = port
        self.scheme = scheme
        self.session = requests.Session()
        self.base_url = '{}://{}:{}/{}'.format(scheme, host, port, self.api_version)

    def register_service(self, name: str = None,
                         service_id=None,
                         tags: List[str] = None,
                         address: str = None,
                         port: int = None,
                         check_http: str = None,
                         check_interval: int = 10):
        api_path = '/agent/service/register?replace-existing-checks=true'
        data = {
            'Name': name,
            'ID': service_id,
            'Tags': tags,
            'Address': address,
            'Port': port,
            'Check': {
                'Name': name,
                'Interval': '{}s'.format(check_interval),
                'HTTP': check_http
            }
        }
        url = '{}{}'.format(self.base_url, api_path)
        resp = self.session.put(url, json=data)
        if not resp.ok:
            raise ConsulClientError(resp.text)

    def deregister_service(self, service_id: str):
        api_path = '/agent/service/deregister/{}'.format(service_id)
        url = '{}{}'.format(self.base_url, api_path)
        resp = self.session.put(url)
        if not resp.ok:
            raise ConsulClientError(resp.text)



# coding=utf-8

from .annotation import *
from .annotation_config_registry import *
from .annotation_config_utils import *
from .bean_context import *
from .bean_name_generator import *
from .condition import *
from .condition_evaluator import *
from .config_constants import *
from .config_post_processor import *
from .context_aware_processor import *
from .default_bean_context import *
from .event import *
from .event_listener import *
from .event_publisher import *

__all__ = (annotation.__all__ +
           annotation_config_registry.__all__ +
           annotation_config_utils.__all__ +
           bean_context.__all__ +
           bean_name_generator.__all__ +
           condition.__all__ +
           condition_evaluator.__all__ +
           config_constants.__all__ +
           config_post_processor.__all__ +
           context_aware_processor.__all__ +
           default_bean_context.__all__ +
           event.__all__ +
           event_listener.__all__ +
           event_publisher.__all__)



# coding=utf-8

from typing import Union
import inspect

from flask import current_app, Flask

from guniflask.security_config.http_security_builder import HttpSecurityBuilder
from guniflask.security_config.authentication_manager_builder import AuthenticationManagerBuilder
from guniflask.security.user_details_service import UserDetailsService
from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.web.request_filter import RequestFilter, RequestFilterChain
from guniflask.security_config.http_basic_configurer import HttpBasicConfigurer
from guniflask.security_config.cors_configurer import CorsConfigurer
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.context.bean_context import BeanContext
from guniflask.annotation.core import AnnotationUtils
from guniflask.web.bind_annotation import Blueprint

__all__ = ['HttpSecurity']


class HttpSecurity(HttpSecurityBuilder):
    def __init__(self, authentication_builder: AuthenticationManagerBuilder,
                 shared_objects: dict = None):
        super().__init__()
        assert authentication_builder is not None, 'Authentication builder is required'
        self.set_shared_object(AuthenticationManagerBuilder, authentication_builder)
        for k, v in shared_objects.items():
            self.set_shared_object(k, v)
        self._security_filter_chain = RequestFilterChain()
        self._blueprints = []

    def _perform_build(self):
        if not self._blueprints:
            self._blueprints = [current_app._get_current_object()]
        for b in self._blueprints:
            if isinstance(b, Flask):
                b.before_request(self._security_filter_chain.before_request)
                b.after_request(self._security_filter_chain.after_request)
            else:
                b_cls = type(b)
                annotation = AnnotationUtils.get_annotation(b_cls, Blueprint)
                if annotation:
                    blueprint = annotation['blueprint']
                    if blueprint:
                        blueprint.before_request(self._security_filter_chain.before_request)
                        blueprint.after_request(self._security_filter_chain.after_request)

    def _before_configure(self):
        self.set_shared_object(AuthenticationManager, self._get_authentication_registry().build())

    def with_user_details_service(self, user_details_service: UserDetailsService) -> 'HttpSecurity':
        self._get_authentication_registry().with_user_details_service(user_details_service)
        return self

    def with_authentication_provider(self, authentication_provider: AuthenticationProvider) -> 'HttpSecurity':
        self._get_authentication_registry().with_authentication_provider(authentication_provider)
        return self

    def _get_authentication_registry(self) -> AuthenticationManagerBuilder:
        return self.get_shared_object(AuthenticationManagerBuilder)

    def http_basic(self) -> HttpBasicConfigurer:
        return self._get_or_apply(HttpBasicConfigurer())

    def cors(self, config) -> CorsConfigurer:
        return self._get_or_apply(CorsConfigurer(config))

    def _get_or_apply(self, configurer):
        existing = self.get_configurer(type(configurer))
        if existing:
            return existing
        return self.apply(configurer)

    def add_request_filter(self, request_filter: RequestFilter) -> 'HttpSecurity':
        self._security_filter_chain.add_request_filter(request_filter)
        return self

    def authorize_blueprint(self, blueprint: Union[str, type]) -> 'HttpSecurity':
        context: BeanContext = self.get_shared_object(BeanContext)
        if inspect.isclass(blueprint):
            bean = context.get_bean_of_type(blueprint)
            if not bean:
                raise ValueError('No such blueprint with type "{}"'.format(blueprint.__name__))
        elif isinstance(blueprint, str):
            bean = context.get_bean(blueprint)
            if not bean:
                raise ValueError('No such blueprint named "{}"'.format(blueprint))
        else:
            raise ValueError('Cannot resolve the blueprint: {}'.format(blueprint))
        self._blueprints.append(bean)
        return self



# coding=utf-8

from flask import Flask

from guniflask.context.default_bean_context import AnnotationConfigBeanContext
from guniflask.web.blueprint_post_processor import BlueprintPostProcessor
from guniflask.security_config.web_security_config import WebSecurityConfiguration
from guniflask.web.scheduling_config import WebAsyncConfiguration, WebSchedulingConfiguration
from guniflask.web.config_constants import *
from guniflask.beans.definition import BeanDefinition

__all__ = ['WebApplicationContext']


class WebApplicationContext(AnnotationConfigBeanContext):
    def __init__(self, app: Flask):
        super().__init__()
        self.app = app

    def _post_process_bean_factory(self):
        super()._post_process_bean_factory()

        if not self.contains_bean_definition(BLUEPRINT_POST_PROCESSOR):
            bean_definition = BeanDefinition(BlueprintPostProcessor)
            self.register_bean_definition(BLUEPRINT_POST_PROCESSOR, bean_definition)

        self._reader.register(WebSecurityConfiguration)
        self._reader.register(WebAsyncConfiguration)
        self._reader.register(WebSchedulingConfiguration)



# coding=utf-8

import logging
from importlib import import_module

from flask import Blueprint

from guniflask.utils.logging import redirect_app_logger, redirect_logger
from guniflask.config.app_config import AppConfig
from guniflask.utils.env import walk_modules
from guniflask.web.context import WebApplicationContext

__all__ = ['AppInitializer']


class AppInitializer:
    def __init__(self, app):
        self.app = app
        self.config = AppConfig(self.app)

    def init(self):
        self._configure_logger()
        self._make_settings()
        bean_context = self._create_bean_context()
        self.app.bean_context = bean_context  # register bean context for app
        self._init_app()
        with self.app.app_context():
            self._register_blueprints()
            self._refresh_bean_context(bean_context)

    def _configure_logger(self):
        """
        Reuse gunicorn logger
        """
        gunicorn_logger = logging.getLogger('gunicorn.error')
        redirect_logger('guniflask', gunicorn_logger)
        redirect_app_logger(self.app, gunicorn_logger)
        redirect_logger(self.app.name, gunicorn_logger)

    def _make_settings(self):
        app_module = self._get_app_module()
        _make_settings = getattr(app_module, 'make_settings', None)
        if _make_settings:
            s = self.config.app_settings(self.app)
            _make_settings(self.app, s)

    def _init_app(self):
        self.config.init_app()
        app_module = self._get_app_module()
        _init_app = getattr(app_module, 'init_app', None)
        if _init_app:
            s = self.config.app_settings(self.app)
            _init_app(self.app, s)

    def _create_bean_context(self) -> WebApplicationContext:
        bean_context = WebApplicationContext(self.app)
        bean_context.scan(self.app.name)
        return bean_context

    def _refresh_bean_context(self, bean_context: WebApplicationContext):
        bean_context.refresh()

    def _register_blueprints(self):
        """
        Register declared Blueprint
        """
        registered_blueprints = set()

        def iter_blueprints():
            for module in walk_modules(self.app.name):
                for obj in vars(module).values():
                    if isinstance(obj, Blueprint) and obj not in registered_blueprints:
                        yield obj
                        registered_blueprints.add(obj)

        for b in iter_blueprints():
            self.app.register_blueprint(b)

        del registered_blueprints

    def _get_app_module(self):
        return import_module(self.app.name + '.app')



# coding=utf-8

from guniflask.security_config.http_security_builder import HttpSecurityBuilder
from guniflask.security_config.security_configurer import SecurityConfigurerAdapter
from guniflask.web.cors import CorsFilter

__all__ = ['CorsConfigurer']


class CorsConfigurer(SecurityConfigurerAdapter):

    def __init__(self, cors=None):
        super().__init__()

        if isinstance(cors, dict):
            self.cors_filter = CorsFilter(**cors)
        else:
            self.cors_filter = CorsFilter()

    def configure(self, http: HttpSecurityBuilder):
        http.add_request_filter(self.cors_filter)



# coding=utf-8

from typing import List, Set

from guniflask.utils.factory import instantiate_from_json


class Person:
    name: str
    age: int


class Teacher(Person):
    classes: Set


class Student(Person):
    secret: str
    mentor: Teacher
    parents: List[Person]
    scores: dict

    def __init__(self):
        self.hobbies: List = None
        self.graduated = False


def test_instantiate_from_json():
    assert instantiate_from_json('1', dtype=int) == 1
    assert instantiate_from_json([1, 2, 3]) == [1, 2, 3]
    assert instantiate_from_json({'key': 'value'}) == {'key': 'value'}

    student_data = dict(
        name='Xiao Ming',
        age=12,
        graduated=True,
        hobbies=['Programming', 'Piano'],
        scores={'Math': 100},
        parents=[
            dict(
                name='Wang Shu',
                age=40
            ),
            dict(
                name='Wang Shener',
                age=39
            )
        ],
        mentor=dict(
            name='Jia Meng',
            age=41,
            classes=['English', 'Yuwen']
        )
    )

    student = instantiate_from_json(student_data, Student)

    assert isinstance(student, Student)
    assert student.name == 'Xiao Ming'
    assert student.age == 12
    assert student.graduated is True
    assert student.hobbies == ['Programming', 'Piano']
    assert student.scores == {'Math': 100}
    assert student.secret is None

    assert isinstance(student.parents, list) and len(student.parents) == 2
    for parent in student.parents:
        if parent.name == 'Wang Shu':
            assert parent.age == 40
        elif parent.name == 'Wang Shener':
            assert parent.age == 39
        else:
            raise RuntimeError

    assert isinstance(student.mentor, Teacher)
    mentor = student.mentor
    assert mentor.name == 'Jia Meng'
    assert mentor.age == 41
    assert isinstance(mentor.classes, set)
    assert mentor.classes == {'English', 'Yuwen'}



# coding=utf-8

import datetime as dt
import uuid

from flask import current_app
from werkzeug.local import LocalProxy

from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.oauth2.errors import InvalidTokenError
from guniflask.oauth2.token import OAuth2AccessToken
from guniflask.oauth2.token_converter import AccessTokenConverter, JwtAccessTokenConverter, \
    UserAuthenticationConverter
from guniflask.security.user import User
from guniflask.security.jwt import JwtHelper

__all__ = ['jwt_manager', 'JwtManager']

jwt_manager = LocalProxy(lambda: current_app.extensions['jwt_manager'])


class JwtManager(AuthenticationManager):
    EXP = AccessTokenConverter.EXP
    JTI = AccessTokenConverter.JTI
    AUTHORITIES = AccessTokenConverter.AUTHORITIES
    USERNAME = UserAuthenticationConverter.USERNAME
    USER_DETAILS = 'user_details'

    def __init__(self, secret='', public_key=None, private_key=None, algorithm='HS256',
                 access_token_expires_in=24 * 60 * 60, refresh_token_expires_in=365 * 24 * 60 * 60):
        self.secret = secret
        self.public_key = public_key
        self.private_key = private_key
        self.algorithm = algorithm
        self.access_token_expires_in = access_token_expires_in
        self.refresh_token_expires_in = refresh_token_expires_in

        self.token_converter = JwtAccessTokenConverter()

        app = current_app._get_current_object()
        app.extensions['jwt_manager'] = self

    def create_access_token(self, authorities=None, username=None, **user_details) -> str:
        expires_in = self.access_token_expires_in
        payload = {
            self.JTI: uuid.uuid4().hex
        }
        exp = dt.datetime.now() + dt.timedelta(seconds=expires_in)
        payload[self.EXP] = exp

        payload[self.AUTHORITIES] = authorities
        payload[self.USERNAME] = username
        payload[self.USER_DETAILS] = user_details

        return JwtHelper.encode_jwt(payload, self.secret or self.private_key,
                                    self.algorithm)

    def read_access_token(self, access_token_value: str) -> OAuth2AccessToken:
        payload = self._decode(access_token_value)
        return self.token_converter.extract_access_token(access_token_value, payload)

    def authenticate(self, authentication):
        access_token_value = authentication.principal
        payload = self._decode(access_token_value)
        authorities = payload[self.AUTHORITIES]
        username = payload[self.USERNAME]
        user_details = payload[self.USER_DETAILS]
        user = User(username=username, authorities=authorities)
        for k, v in user_details.items():
            setattr(user, k, v)
        user_auth = UserAuthentication(user, authorities=authorities)
        user_auth.authenticate(True)
        return user_auth

    def _decode(self, access_token_value):
        try:
            payload = JwtHelper.decode_jwt(access_token_value, self.secret or self.public_key,
                                           self.algorithm)
        except Exception as e:
            raise InvalidTokenError(e)
        return payload



# coding=utf-8

from typing import Optional

from flask import _request_ctx_stack
from werkzeug.local import LocalProxy

from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.context import SecurityContext
from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.security.user import User
from guniflask.security.user_details import UserDetails

__all__ = ['current_user']


def _load_user() -> Optional[User]:
    ctx = _request_ctx_stack.top
    if ctx is not None:
        if not hasattr(ctx, 'user'):
            user = None
            auth = SecurityContext.get_authentication()
            if auth is not None:
                if isinstance(auth, OAuth2Authentication):
                    auth = auth.user_authentication
                if isinstance(auth, UserAuthentication):
                    if isinstance(auth.principal, UserDetails):
                        user = auth.principal
                    else:
                        user = User(username=auth.name, authorities=auth.authorities)
            ctx.user = user
        return ctx.user


current_user = LocalProxy(_load_user)



# coding=utf-8

from guniflask.security.user_details import UserDetails

__all__ = ['User']


class User(UserDetails):
    default_role_prefix = 'role_'

    def __init__(self, username=None, password=None, authorities=None):
        self._username = username
        self._password = password
        self._authorities = set()
        if authorities:
            self._authorities.update(authorities)

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @property
    def authorities(self):
        return self._authorities

    def has_authority(self, authority):
        return self.has_any_authority(authority)

    def has_any_authority(self, *authorities):
        return self._has_any_authority_name(None, *authorities)

    def has_role(self, role):
        return self.has_any_role(role)

    def has_any_role(self, *roles):
        return self._has_any_authority_name(self.default_role_prefix, *roles)

    def _has_any_authority_name(self, prefix, *roles):
        for role in roles:
            r = self._role_with_prefix(prefix, role)
            if r in self.authorities:
                return True
        return False

    @staticmethod
    def _role_with_prefix(prefix, role):
        if role is None or prefix is None:
            return role
        if role.startswith(prefix):
            return role
        return prefix + role



# coding=utf-8

import datetime as dt
from typing import Optional
import uuid
from abc import ABCMeta, abstractmethod

from guniflask.oauth2.authentication import OAuth2Authentication
from guniflask.oauth2.token import OAuth2AccessToken, OAuth2RefreshToken
from guniflask.oauth2.token_store import TokenStore
from guniflask.oauth2.token_converter import TokenEnhancer
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.request import TokenRequest, OAuth2Request
from guniflask.oauth2.errors import InvalidTokenError, InvalidGrantError, InvalidScopeError, ClientRegistrationError
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.preauth_token import PreAuthenticatedToken
from guniflask.beans.factory_hook import InitializingBean

__all__ = ['AuthorizationServerTokenServices',
           'ResourceServerTokenServices',
           'ConsumerTokenServices',
           'DefaultTokenServices']


class AuthorizationServerTokenServices(metaclass=ABCMeta):
    @abstractmethod
    def create_access_token(self, authentication: OAuth2Authentication = None) -> OAuth2AccessToken:
        pass

    @abstractmethod
    def refresh_access_token(self, refresh_token_value: str, token_request: TokenRequest) -> OAuth2AccessToken:
        pass

    @abstractmethod
    def get_access_token(self, authentication) -> OAuth2AccessToken:
        pass


class ResourceServerTokenServices(metaclass=ABCMeta):
    @abstractmethod
    def load_authentication(self, access_token_value: str) -> OAuth2Authentication:
        pass

    @abstractmethod
    def read_access_token(self, access_token_value: str) -> OAuth2AccessToken:
        pass


class ConsumerTokenServices(metaclass=ABCMeta):
    @abstractmethod
    def revoke_token(self, access_token_value: str):
        pass


class DefaultTokenServices(AuthorizationServerTokenServices, ResourceServerTokenServices, ConsumerTokenServices,
                           InitializingBean):
    """
    Base implementation for token services using random UUID values for the access token and refresh token values.
    """

    def __init__(self, token_store: TokenStore):
        self.token_store = token_store
        self.token_enhancer: TokenEnhancer = None
        self.client_details_service: ClientDetailsService = None
        self.access_token_expires_in = 24 * 60 * 60
        self.refresh_token_expires_in = 365 * 24 * 60 * 60
        self.support_refresh_token = False
        self.authentication_manager: AuthenticationManager = None

    def after_properties_set(self):
        assert self.token_store is not None, 'Token store must be set'

    def create_access_token(self, authentication: OAuth2Authentication = None) -> OAuth2AccessToken:
        existing_access_token = self.token_store.get_access_token(authentication)
        refresh_token = None
        if existing_access_token is not None:
            if existing_access_token.is_expired:
                if existing_access_token.refresh_token is not None:
                    refresh_token = existing_access_token.refresh_token
                    self.token_store.remove_refresh_token(refresh_token)
                self.token_store.remove_access_token(existing_access_token)
            else:
                self.token_store.store_access_token(existing_access_token, authentication)
                return existing_access_token

        if refresh_token is None:
            refresh_token = self._create_refresh_token(authentication)
        elif refresh_token.is_expired:
            refresh_token = self._create_refresh_token(authentication)

        access_token = self._do_create_access_token(authentication, refresh_token)
        self.token_store.store_access_token(access_token, authentication)
        if access_token.refresh_token is not None:
            self.token_store.store_refresh_token(refresh_token, authentication)
        return access_token

    def refresh_access_token(self, refresh_token_value: str, token_request: TokenRequest) -> OAuth2AccessToken:
        if not self.support_refresh_token:
            raise InvalidGrantError('Invalid refresh token: {}'.format(refresh_token_value))
        refresh_token = self.token_store.read_refresh_token(refresh_token_value)
        if refresh_token is None:
            raise InvalidGrantError('Invalid refresh token: {}'.format(refresh_token_value))

        authentication = self.token_store.read_authentication_for_refresh_token(refresh_token)
        if self.authentication_manager is not None and not authentication.is_client_only:
            user = PreAuthenticatedToken(authentication.user_authentication, authorities=authentication.authorities)
            user = self.authentication_manager.authenticate(user)
            details = authentication.details
            authentication = OAuth2Authentication(authentication.oauth2_request, user)
            authentication.details = details
        client_id = authentication.oauth2_request.client_id
        if client_id is None or client_id != token_request.client_id:
            raise InvalidGrantError('Wrong client for this refresh token: {}'.format(refresh_token_value))

        self.token_store.remove_access_token_using_refresh_token(refresh_token)
        if refresh_token.is_expired:
            self.token_store.remove_refresh_token(refresh_token)
            raise InvalidTokenError('Refresh token expired: {}'.format(refresh_token_value))
        authentication = self._create_refreshed_authentication(authentication, token_request)
        access_token = self._do_create_access_token(authentication, refresh_token)
        self.token_store.store_access_token(access_token, authentication)
        return access_token

    def get_access_token(self, authentication) -> OAuth2AccessToken:
        return self.token_store.get_access_token(authentication)

    def load_authentication(self, access_token_value: str) -> OAuth2Authentication:
        access_token = self.token_store.read_access_token(access_token_value)
        if access_token is None:
            raise InvalidTokenError('Invalid access token: {}'.format(access_token_value))
        elif access_token.is_expired:
            self.token_store.remove_access_token(access_token)
            raise InvalidTokenError('Access token expired: {}'.format(access_token_value))

        result = self.token_store.read_authentication(access_token)
        if result is None:
            raise InvalidTokenError('Invalid access token: {}'.format(access_token_value))
        if self.client_details_service is not None:
            client_id = result.oauth2_request.client_id
            try:
                self.client_details_service.load_client_details_by_client_id(client_id)
            except ClientRegistrationError:
                raise InvalidTokenError('Client not valid: {}'.format(client_id))
        return result

    def read_access_token(self, access_token_value: str) -> OAuth2AccessToken:
        return self.token_store.read_access_token(access_token_value)

    def revoke_token(self, access_token_value: str):
        access_token = self.token_store.read_access_token(access_token_value)
        if access_token:
            if access_token.refresh_token:
                self.token_store.remove_refresh_token(access_token.refresh_token)
            self.token_store.remove_access_token(access_token)

    def _create_refresh_token(self, authentication: OAuth2Authentication) -> Optional[OAuth2RefreshToken]:
        if not self._is_support_refresh_token(authentication.oauth2_request):
            return None
        expires_in = self._get_refresh_token_expires_in(authentication.oauth2_request)
        value = uuid.uuid4().hex
        if expires_in <= 0:
            expiration = None
        else:
            expiration = dt.datetime.now() + dt.timedelta(seconds=expires_in)
        return OAuth2RefreshToken(value, expiration=expiration)

    def _is_support_refresh_token(self, client_auth: OAuth2Request) -> bool:
        if self.client_details_service is not None:
            client = self.client_details_service.load_client_details_by_client_id(client_auth.client_id)
            return 'refresh_token' in client.grant_types
        return self.support_refresh_token

    def _get_refresh_token_expires_in(self, client_auth: OAuth2Request) -> int:
        if self.client_details_service is not None:
            client = self.client_details_service.load_client_details_by_client_id(client_auth.client_id)
            expires_in = client.refresh_token_expires_in
            if expires_in is not None:
                return expires_in
        return self.refresh_token_expires_in

    def _do_create_access_token(self, authentication: OAuth2Authentication,
                                refresh_token: OAuth2RefreshToken) -> OAuth2AccessToken:
        token = OAuth2AccessToken(uuid.uuid4().hex)
        expires_in = self._get_access_token_expires_in(authentication.oauth2_request)
        if expires_in <= 0:
            expiration = None
        else:
            expiration = dt.datetime.now() + dt.timedelta(seconds=expires_in)
        token.expiration = expiration
        token.refresh_token = refresh_token
        token.scope = authentication.oauth2_request.scope

        if self.token_enhancer is not None:
            return self.token_enhancer.enhance(token, authentication)
        return token

    def _get_access_token_expires_in(self, client_auth: OAuth2Request) -> int:
        if self.client_details_service is not None:
            client = self.client_details_service.load_client_details_by_client_id(client_auth.client_id)
            expires_in = client.access_token_expires_in
            if expires_in is not None:
                return expires_in
        return self.access_token_expires_in

    def _create_refreshed_authentication(self, authentication: OAuth2Authentication,
                                         request: TokenRequest) -> OAuth2Authentication:
        scope = request.scope
        client_auth = authentication.oauth2_request.do_refresh(request)
        if scope:
            original_scope = client_auth.scope
            if original_scope is None or not original_scope.issuperset(scope):
                raise InvalidScopeError('Unable to narrow the scope of the client authentication to: {}'.format(scope))
            else:
                client_auth = client_auth.narrow_scope(scope)
        narrowed = OAuth2Authentication(client_auth, authentication.user_authentication)
        return narrowed



# coding=utf-8

from abc import ABCMeta, abstractmethod

from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.oauth2.token_service import ResourceServerTokenServices, DefaultTokenServices
from guniflask.oauth2.token_store import TokenStore
from guniflask.oauth2.token_extractor import TokenExtractor
from guniflask.oauth2.authentication_manager import OAuth2AuthenticationManager
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.security_config.security_configurer import SecurityConfigurerAdapter
from guniflask.oauth2.authentication_filter import OAuth2AuthenticationFilter
from guniflask.security_config.http_security import HttpSecurity

__all__ = ['ResourceServerConfigurer',
           'ResourceServerConfigurerAdapter',
           'ResourceServerSecurityConfigurer']


class ResourceServerSecurityConfigurer(SecurityConfigurerAdapter):
    def __init__(self):
        super().__init__()
        self._authentication_manager: AuthenticationManager = None
        self._resource_token_services: ResourceServerTokenServices = None
        self._token_store: TokenStore = None
        self._resource_id: str = 'oauth2-resource'
        self._token_extractor: TokenExtractor = None

    @property
    def token_store(self) -> TokenStore:
        return self._token_store

    def with_token_store(self, token_store: TokenStore) -> 'ResourceServerSecurityConfigurer':
        self._token_store = token_store
        return self

    def with_token_extractor(self, token_extractor: TokenExtractor) -> 'ResourceServerSecurityConfigurer':
        self._token_extractor = token_extractor
        return self

    def with_authentication_manager(self, authentication_manager: AuthenticationManager
                                    ) -> 'ResourceServerSecurityConfigurer':
        self._authentication_manager = authentication_manager
        return self

    def with_token_services(self, token_services: ResourceServerTokenServices) -> 'ResourceServerSecurityConfigurer':
        self._resource_token_services = token_services
        return self

    def with_resource_id(self, resource_id: str):
        self._resource_id = resource_id
        return self

    def configure(self, http: HttpSecurity):
        authentication_manager = self._get_oauth2_authentication_manager()
        authentication_filter = OAuth2AuthenticationFilter()
        authentication_filter.authentication_manager = authentication_manager
        if self._token_extractor is not None:
            authentication_filter.token_extractor = self._token_extractor
        http.add_request_filter(authentication_filter)

    def _get_oauth2_authentication_manager(self) -> AuthenticationManager:
        if self._authentication_manager:
            if isinstance(self._authentication_manager, OAuth2AuthenticationManager):
                authentication_manager = self._authentication_manager
            else:
                return self._authentication_manager
        else:
            authentication_manager = OAuth2AuthenticationManager()

        authentication_manager.resource_id = self._resource_id
        authentication_manager.token_services = self._get_resource_token_services()
        authentication_manager.client_details_service = self._get_client_details_service()
        return authentication_manager

    def _get_resource_token_services(self) -> ResourceServerTokenServices:
        if self._resource_token_services is None:
            token_services = DefaultTokenServices(self._get_token_store())
            token_services.support_refresh_token = True
            token_services.client_details_service = self._get_client_details_service()
            self.resource_token_services = token_services
        return self.resource_token_services

    def _get_client_details_service(self) -> ClientDetailsService:
        http: HttpSecurity = self.builder
        return http.get_shared_object(ClientDetailsService)

    def _get_token_store(self):
        assert self.token_store is not None, 'Token store is required'
        return self.token_store


class ResourceServerConfigurer(metaclass=ABCMeta):
    @abstractmethod
    def configure_security(self, resources: ResourceServerSecurityConfigurer):
        pass

    @abstractmethod
    def configure_http(self, http: HttpSecurity):
        pass


class ResourceServerConfigurerAdapter(ResourceServerConfigurer):
    def configure_security(self, resources: ResourceServerSecurityConfigurer):
        pass

    def configure_http(self, http: HttpSecurity):
        pass



# coding=utf-8

import sys
import tty
import termios


def readchar():  # pragma: no cover
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch



# coding=utf-8

from importlib import import_module
from typing import get_type_hints, List, Set, Any, Mapping, MutableMapping, Optional
import datetime as dt
import inspect
from collections import OrderedDict


def load_object(path):
    if isinstance(path, str):
        dot = path.rindex('.')
        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)
        return getattr(mod, name)
    return path


def string_to_datetime(s):
    if isinstance(s, int):
        return dt.datetime.fromtimestamp(s, tz=dt.timezone.utc).astimezone()
    elif isinstance(s, str):
        if 'GMT' in s:
            return dt.datetime.strptime(s, '%a, %d %b %Y %H:%M:%S GMT').replace(tzinfo=dt.timezone.utc).astimezone()
    return s


def instantiate_from_json(source, dtype: Any = None, target=None) -> Any:
    if source is None:
        return
    if isinstance(source, Mapping):
        if dtype is None:
            # inference type from target
            if target is not None:
                dtype = type(target)
        if dtype is None:
            # cannot instantiate without type
            return source

        type_hints = {} if dtype is None else get_type_hints(dtype)
        if target is None:
            target = dtype()

        if isinstance(target, MutableMapping):
            for k, v in source.items():
                target[k] = v
        else:
            for key in source:
                value = source[key]
                prop = getattr(target, key, None)
                prop_type = type_hints.get(key)

                prop_value = instantiate_from_json(value, dtype=prop_type, target=prop)
                if prop_value is not None:
                    setattr(target, key, prop_value)

            for key in type_hints:
                if not hasattr(target, key):
                    setattr(target, key, None)
        return target
    elif isinstance(source, List):
        if dtype is None:
            return source

        argc, etype = resolve_arg_type(dtype)

        if argc is ArgType.LIST:
            collection_type = list
        elif argc is ArgType.SET:
            collection_type = set
        else:
            assert inspect.isclass(dtype), 'Cannot convert a list to {}'.format(dtype)
            collection_type = dtype

        result = []
        for v in source:
            result.append(instantiate_from_json(v, dtype=etype))
        if collection_type is not None:
            result = collection_type(result)
        return result
    else:
        if dtype == dt.datetime:
            source = string_to_datetime(source)
        elif dtype is not None:
            try:
                source = dtype(source)
            except Exception:
                source = None
        return source


def inspect_args(func):
    signature = inspect.signature(func)
    paramters = signature.parameters
    return_type = signature.return_annotation

    hints = {}
    if return_type is not inspect._empty:
        hints['return'] = return_type

    args = OrderedDict()
    for p in paramters.values():
        args[p.name] = p.default
        if p.annotation is not inspect._empty:
            hints[p.name] = p.annotation

    return args, hints


def resolve_arg_type(arg_type: Optional[type]):
    if arg_type is None:
        return ArgType.SINGLE, arg_type

    if hasattr(arg_type, '__origin__'):
        # handle generic type
        origin = getattr(arg_type, '__origin__')
        if origin is None:
            origin = arg_type

        argc = None
        etype = None
        if issubclass(origin, List):
            argc = ArgType.LIST
            if hasattr(arg_type, '__args__'):
                args = getattr(arg_type, '__args__')
                if args and len(args) == 1:
                    vt = args[0]
                    if not inspect.isclass(vt):
                        vt = None
                    etype = vt
        elif issubclass(origin, Mapping):
            argc = ArgType.DICT
            if hasattr(arg_type, '__args__'):
                args = getattr(arg_type, '__args__')
                if args and len(args) == 2:
                    kt = args[0]
                    vt = args[1]
                    if not inspect.isclass(kt):
                        kt = None
                    if not inspect.isclass(vt):
                        vt = None
                    etype = (kt, vt)
        elif issubclass(origin, Set):
            argc = ArgType.SET
            if hasattr(arg_type, '__args__'):
                args = getattr(arg_type, '__args__')
                if args and len(args) == 1:
                    vt = args[0]
                    if not inspect.isclass(vt):
                        vt = None
                    etype = vt
        if argc is None:
            raise ValueError('Unsupported generic argument type: {}'.format(arg_type))
        return argc, etype
    else:
        assert inspect.isclass(arg_type), 'Non-generic argument type must be a class, but got: {}'.format(arg_type)
        if issubclass(arg_type, List):
            return ArgType.LIST, None
        if issubclass(arg_type, Mapping):
            return ArgType.DICT, None
        if issubclass(arg_type, Set):
            return ArgType.SET, None
        return ArgType.SINGLE, arg_type


class ArgType:
    SINGLE = object()
    DICT = object()
    LIST = object()
    SET = object()



# coding=utf-8

from typing import List

from guniflask.context.annotation import configuration, bean, include
from guniflask.beans.annotation import autowired
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.token_converter import JwtAccessTokenConverter
from guniflask.oauth2.token_endpoint import TokenKeyEndpoint, TokenEndpoint
from guniflask.oauth2_config.client_details_service_configurer import ClientDetailsServiceConfigurer
from guniflask.oauth2_config.client_details_service_config import ClientDetailsServiceConfiguration
from guniflask.oauth2_config.authorization_server_configurer import AuthorizationServerConfigurer, \
    AuthorizationServerEndpointsConfigurer, AuthorizationServerSecurityConfigurer
from guniflask.security_config.web_security_configurer import WebSecurityConfigurer
from guniflask.security_config.authentication_manager_builder import AuthenticationManagerBuilder
from guniflask.security_config.http_security import HttpSecurity

__all__ = ['AuthorizationServerEndpointsConfiguration',
           'AuthorizationServerSecurityConfiguration']


@configuration
class AuthorizationServerEndpointsConfiguration:
    def __init__(self, client_details_service: ClientDetailsService,
                 configurers: List[AuthorizationServerConfigurer] = None):
        self._client_details_service = client_details_service
        self._endpoints = AuthorizationServerEndpointsConfigurer()
        self._endpoints.with_client_details_service(client_details_service)
        if configurers:
            for c in configurers:
                c.configure_endpoints(self._endpoints)

    @property
    def endpoints_configurer(self):
        return self._endpoints

    @bean
    def token_endpoint(self) -> TokenEndpoint:
        token_endpoint = TokenEndpoint(self._endpoints.token_granter, self._client_details_service)
        token_endpoint.oauth2_request_factory = self._endpoints.oauth2_request_factory
        token_endpoint.oauth2_request_validator = self._endpoints.oauth2_request_validator
        return token_endpoint

    @bean
    def token_key_endpoint(self) -> TokenKeyEndpoint:
        access_token_converter = self._endpoints.access_token_converter
        if isinstance(access_token_converter, JwtAccessTokenConverter):
            token_key_endpoint = TokenKeyEndpoint(access_token_converter)
            return token_key_endpoint


@configuration
@include(ClientDetailsServiceConfiguration, AuthorizationServerEndpointsConfiguration)
class AuthorizationServerSecurityConfiguration(WebSecurityConfigurer):
    def __init__(self, client_details_service: ClientDetailsService,
                 endpoints: AuthorizationServerEndpointsConfiguration,
                 configurers: List[AuthorizationServerConfigurer] = None):
        super().__init__()
        self._client_details_service = client_details_service
        self._endpoints = endpoints
        self._configurers = configurers

    @autowired
    def configure_client_details_service(self, client_details: ClientDetailsServiceConfigurer):
        if self._configurers:
            for c in self._configurers:
                c.configure_client_details_service(client_details)

    def configure_http(self, http: HttpSecurity):
        configurer = AuthorizationServerSecurityConfigurer()
        self._configure_security(configurer)
        http.apply(configurer)

        http \
            .authorize_blueprint(TokenEndpoint) \
            .authorize_blueprint(TokenKeyEndpoint) \
            .http_basic()

        http.set_shared_object(ClientDetailsService, self._client_details_service)

    def configure_authentication(self, authentication_builder: AuthenticationManagerBuilder):
        self._enable_local_authentication = True

    def _configure_security(self, configurer: AuthorizationServerSecurityConfigurer):
        if self._configurers:
            for c in self._configurers:
                c.configure_security(configurer)



# coding=utf-8

from abc import ABCMeta, abstractmethod
from typing import Any

__all__ = ['BeanPostProcessor', 'BeanPostProcessorAdapter']


class BeanPostProcessor(metaclass=ABCMeta):
    @abstractmethod
    def post_process_before_instantiation(self, bean_type: type, bean_name: str):
        pass

    @abstractmethod
    def post_process_before_initialization(self, bean: Any, bean_name: str) -> Any:
        pass

    @abstractmethod
    def post_process_after_initialization(self, bean: Any, bean_name: str) -> Any:
        pass


class BeanPostProcessorAdapter(BeanPostProcessor):
    def post_process_before_instantiation(self, bean_type: type, bean_name: str):
        return None

    def post_process_before_initialization(self, bean, bean_name: str):
        return bean

    def post_process_after_initialization(self, bean, bean_name: str):
        return bean



# coding=utf-8

from typing import List

from guniflask.annotation.core import Annotation, AnnotationUtils

__all__ = ['FilterChain', 'filter_chain',
           'MethodFilter',
           'before_request', 'after_request',
           'app_before_request', 'app_after_request',
           'error_handler', 'app_error_handler']


class FilterChain(Annotation):
    def __init__(self, values: List = None):
        super().__init__(values=values)


def filter_chain(*values):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, FilterChain(list(values)))
        return func

    return wrap_func


class MethodFilter(Annotation):
    def add_method_filter(self, name, args=None):
        if 'values' not in self:
            self['values'] = []
        self['values'].append({'name': name, 'args': args})

    def get_method_filters(self):
        if 'values' not in self:
            return []
        return self['values']


def _add_method_filter_annotation(func, name, args=None):
    a = AnnotationUtils.get_annotation(func, MethodFilter)
    if a is None:
        a = MethodFilter()
        AnnotationUtils.add_annotation(func, a)
    a.add_method_filter(name, args=args)


def before_request(func):
    _add_method_filter_annotation(func, 'before_request')
    return func


def after_request(func):
    _add_method_filter_annotation(func, 'after_request')
    return func


def app_before_request(func):
    _add_method_filter_annotation(func, 'app_before_request')
    return func


def app_after_request(func):
    _add_method_filter_annotation(func, 'app_after_request')
    return func


def error_handler(code_or_exception):
    def wrap_func(func):
        _add_method_filter_annotation(func, 'errorhandler', args=[code_or_exception])
        return func

    return wrap_func


def app_error_handler(code_or_exception):
    def wrap_func(func):
        _add_method_filter_annotation(func, 'app_errorhandler', args=[code_or_exception])
        return func

    return wrap_func



# coding=utf-8

from functools import update_wrapper

from flask import current_app, _request_ctx_stack, copy_current_request_context

from guniflask.context.annotation import bean, configuration
from guniflask.scheduling.config_constants import ASYNC_ANNOTATION_PROCESSOR, SCHEDULED_ANNOTATION_PROCESSOR
from guniflask.scheduling.async_config import AsyncConfiguration
from guniflask.scheduling.async_post_processor import AsyncPostProcessor
from guniflask.scheduling.scheduling_config import SchedulingConfiguration
from guniflask.scheduling.scheduling_post_processor import ScheduledPostProcessor

__all__ = ['WebAsyncConfiguration', 'WebAsyncPostProcessor',
           'WebSchedulingConfiguration', 'WebScheduledPostProcessor']


@configuration
class WebAsyncConfiguration(AsyncConfiguration):

    @bean(ASYNC_ANNOTATION_PROCESSOR)
    def async_annotation_processor(self) -> AsyncPostProcessor:
        return WebAsyncPostProcessor()


class WebAsyncPostProcessor(AsyncPostProcessor):

    def _post_process_async_method(self, method):
        wrapped_method = run_with_context(method)
        return super()._post_process_async_method(wrapped_method)


@configuration
class WebSchedulingConfiguration(SchedulingConfiguration):

    @bean(SCHEDULED_ANNOTATION_PROCESSOR)
    def scheduled_annotation_processor(self) -> ScheduledPostProcessor:
        return WebScheduledPostProcessor()


class WebScheduledPostProcessor(ScheduledPostProcessor):

    def _post_process_scheduled_method(self, method):
        wrapped_method = run_with_context(method)
        return super()._post_process_scheduled_method(wrapped_method)


def run_with_context(func):
    if _request_ctx_stack.top is not None:
        func = copy_current_request_context(func)
    app = current_app._get_current_object()

    def wrapper(*args, **kwargs):
        if app is not None:
            with app.app_context():
                return func(*args, **kwargs)
        return func(*args, **kwargs)

    return update_wrapper(wrapper, func)



# coding=utf-8

from .base_model import *
from .model_utils import *
from .sqlalchemy_wrapper import *

__all__ = (base_model.__all__ +
           model_utils.__all__ +
           sqlalchemy_wrapper.__all__)



# coding=utf-8

from .annotation import *
from .authentication_config import *
from .authentication_manager_builder import *
from .configured_security_builder import *
from .cors_configurer import *
from .dao_authentication_configurer import *
from .http_basic_configurer import *
from .http_security import *
from .http_security_builder import *
from .provider_manager_builder import *
from .security_builder import *
from .security_configurer import *
from .user_details_aware_configurer import *
from .web_security import *
from .web_security_config import *
from .web_security_configurer import *

__all__ = (annotation.__all__ +
           authentication_config.__all__ +
           authentication_manager_builder.__all__ +
           configured_security_builder.__all__ +
           cors_configurer.__all__ +
           dao_authentication_configurer.__all__ +
           http_basic_configurer.__all__ +
           http_security.__all__ +
           http_security_builder.__all__ +
           provider_manager_builder.__all__ +
           security_builder.__all__ +
           security_configurer.__all__ +
           user_details_aware_configurer.__all__ +
           web_security.__all__ +
           web_security_config.__all__ +
           web_security_configurer.__all__)



# coding=utf-8

from abc import ABCMeta, abstractmethod

from flask import request, g

from guniflask.security.authentication import Authentication
from guniflask.security.preauth_token import PreAuthenticatedToken
from guniflask.oauth2.token import OAuth2AccessToken
from guniflask.oauth2.authentication_details import OAuth2AuthenticationDetails

__all__ = ['TokenExtractor', 'BearerTokenExtractor']


class TokenExtractor(metaclass=ABCMeta):
    @abstractmethod
    def extract(self) -> Authentication:
        pass


class BearerTokenExtractor(TokenExtractor):
    def extract(self) -> Authentication:
        token = self._extract_token_from_header()
        if token is None:
            token = self._extract_token_from_query()
        if token is not None:
            return PreAuthenticatedToken(token)

    @staticmethod
    def _extract_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.lower().startswith(OAuth2AccessToken.BEARER_TYPE.lower()):
            g.setdefault(OAuth2AuthenticationDetails.ACCESS_TOKEN_TYPE, OAuth2AccessToken.BEARER_TYPE)
            return auth.split(' ', 1)[1]

    @staticmethod
    def _extract_token_from_query():
        token = request.args.get(OAuth2AccessToken.ACCESS_TOKEN)
        if token is not None:
            g.setdefault(OAuth2AuthenticationDetails.ACCESS_TOKEN_TYPE, OAuth2AccessToken.BEARER_TYPE)



# coding=utf-8

from abc import ABCMeta, abstractmethod
from typing import Union

from guniflask.oauth2.client_details import ClientDetails
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.oauth2_utils import OAuth2Utils
from guniflask.oauth2.request import AuthorizationRequest, OAuth2Request, TokenRequest
from guniflask.oauth2.errors import InvalidClientError, InvalidScopeError

__all__ = ['OAuth2RequestFactory', 'DefaultOAuth2RequestFactory',
           'OAuth2RequestValidator', 'DefaultOAuth2RequestValidator']


class OAuth2RequestFactory(metaclass=ABCMeta):
    @abstractmethod
    def create_authorization_request(self, authorization_parameters: dict) -> AuthorizationRequest:
        pass

    @abstractmethod
    def create_oauth2_request_from_authorization_request(self, request: AuthorizationRequest) -> OAuth2Request:
        pass

    @abstractmethod
    def create_oauth2_request_from_token_request(self, request: TokenRequest,
                                                 client_details: ClientDetails) -> OAuth2Request:
        pass

    @abstractmethod
    def create_token_request(self, request_parameters: dict, authenticated_client: ClientDetails) -> TokenRequest:
        pass

    @abstractmethod
    def create_token_request_from_authorization_request(self, request: AuthorizationRequest,
                                                        grant_type: str) -> TokenRequest:
        pass


class DefaultOAuth2RequestFactory(OAuth2RequestFactory):
    """
    Strategy for managing OAuth2 requests
    """

    def __init__(self, client_details_service: ClientDetailsService):
        self.client_details_service = client_details_service

    def create_authorization_request(self, authorization_parameters: dict):
        client_id = authorization_parameters.get(OAuth2Utils.CLIENT_ID)
        state = authorization_parameters.get(OAuth2Utils.STATE)
        redirect_uri = authorization_parameters.get(OAuth2Utils.REDIRECT_URI)
        response_types = OAuth2Utils.parse_parameter_list(authorization_parameters.get(OAuth2Utils.RESPONSE_TYPE))
        scopes = self._extract_scopes(authorization_parameters, client_id)
        request = AuthorizationRequest(request_parameters=authorization_parameters,
                                       client_id=client_id,
                                       scope=scopes,
                                       approved=False,
                                       state=state,
                                       redirect_uri=redirect_uri,
                                       response_types=response_types)
        client_details = self.client_details_service.load_client_details_by_client_id(client_id)
        request.set_from_client_details(client_details)
        return request

    def create_oauth2_request_from_authorization_request(self, request: AuthorizationRequest):
        oauth2_request = OAuth2Request(request_parameters=request.request_parameters,
                                       client_id=request.client_id,
                                       scope=request.scope,
                                       authorities=request.authorities,
                                       approved=request.approved,
                                       resource_ids=request.resource_ids,
                                       redirect_uri=request.redirect_uri,
                                       response_types=request.response_types)
        return oauth2_request

    def create_token_request(self, request_parameters: dict, authenticated_client: ClientDetails):
        client_id = request_parameters.get(OAuth2Utils.CLIENT_ID)
        if client_id is None:
            client_id = authenticated_client.client_id
        else:
            if client_id != authenticated_client.client_id:
                raise InvalidClientError('Given client ID does not match authenticated client')
        grant_type = request_parameters.get(OAuth2Utils.GRANT_TYPE)
        scopes = self._extract_scopes(request_parameters, client_id)
        request = TokenRequest(request_parameters=request_parameters,
                               client_id=client_id,
                               scope=scopes,
                               grant_type=grant_type)
        return request

    def create_token_request_from_authorization_request(self, request: AuthorizationRequest, grant_type: str):
        token_request = TokenRequest(request_parameters=request.request_parameters,
                                     client_id=request.client_id,
                                     scope=request.scope,
                                     grant_type=grant_type)
        return token_request

    def create_oauth2_request_from_token_request(self, request: TokenRequest, client_details: ClientDetails):
        request_parameters = dict(request.request_parameters)
        if 'password' in request_parameters:
            request_parameters.pop('password')
        if 'client_secret' in request_parameters:
            request_parameters.pop('client_secret')
        request_parameters['grant_type'] = request.grant_type
        oauth2_request = OAuth2Request(request_parameters=request_parameters,
                                       client_id=client_details.client_id,
                                       scope=request.scope,
                                       authorities=client_details.authorities,
                                       approved=True,
                                       resource_ids=client_details.resource_ids)
        return oauth2_request

    def _extract_scopes(self, request_parameters: dict, client_id: str):
        scopes = OAuth2Utils.parse_parameter_list(request_parameters.get(OAuth2Utils.SCOPE))
        client_details = self.client_details_service.load_client_details_by_client_id(client_id)
        if not scopes:
            scopes = client_details.scope
        return scopes


class OAuth2RequestValidator(metaclass=ABCMeta):
    @abstractmethod
    def validate_scope(self, request: Union[AuthorizationRequest, TokenRequest], client: ClientDetails):
        pass


class DefaultOAuth2RequestValidator(OAuth2RequestValidator):
    def validate_scope(self, request: Union[AuthorizationRequest, TokenRequest], client: ClientDetails):
        self._validate_request_scope(request.scope, client.scope)

    def _validate_request_scope(self, request_scopes: set, client_scopes: set):
        if client_scopes:
            for scope in request_scopes:
                if scope not in client_scopes:
                    raise InvalidScopeError('Invalid scope: {}'.format(scope))



# coding=utf-8

from flask_cors.core import get_cors_options, parse_resources
from flask_cors.extension import make_after_request_function

from guniflask.web.request_filter import RequestFilter

__all__ = ['CorsFilter']


class CorsFilter(RequestFilter):
    DEFAULT_CONFIG = dict(origins='*',
                          methods=['GET', 'HEAD', 'POST', 'OPTIONS', 'PUT', 'PATCH', 'DELETE'],
                          allow_headers='*',
                          expose_headers=None,
                          supports_credentials=False,
                          max_age=None)

    def __init__(self, resources=None, **kwargs):
        self.default_config = dict(self.DEFAULT_CONFIG)
        self.resources = {}
        self._after_request = None

        self._set_default_config(**kwargs)
        if resources and isinstance(resources, dict):
            for k, v in resources.items():
                if isinstance(v, dict):
                    self._add_resource(k, **v)
                else:
                    self._add_resource(k)
        self._init()

    def _set_default_config(self, **kwargs):
        for k in self.default_config:
            if k in kwargs:
                self.default_config[k] = kwargs[k]

    def _add_resource(self, pattern, **kwargs):
        self.resources[pattern] = dict(self.DEFAULT_CONFIG)
        for k in self.DEFAULT_CONFIG:
            if k in kwargs:
                self.resources[pattern][k] = kwargs[k]

    def _init(self):
        # fake app
        app = object()
        kwargs = dict(self.default_config)
        if self.resources:
            kwargs['resources'] = self.resources

        options = get_cors_options(app, kwargs)
        resources = parse_resources(options.get('resources'))
        resources = [
            (pattern, get_cors_options(app, options, opts))
            for (pattern, opts) in resources
        ]
        self._after_request = make_after_request_function(resources)

    def after_request(self, response):
        return self._after_request(response)



# coding=utf-8

from functools import update_wrapper

from werkzeug.exceptions import Unauthorized

from guniflask.security.user_context import current_user

__all__ = ['login_required', 'has_any_role', 'has_role', 'has_any_authority', 'has_authority']


def login_required(func):
    def wrapper(*args, **kwargs):
        user = current_user._get_current_object()
        if user is None:
            raise Unauthorized
        return func(*args, **kwargs)

    return update_wrapper(wrapper, func)


def has_any_role(*roles):
    def decorator(func):
        def wrapper(*args, **kwargs):
            user = current_user._get_current_object()
            if user is None or not user.has_any_role(*roles):
                raise Unauthorized
            return func(*args, **kwargs)

        return update_wrapper(wrapper, func)

    return decorator


def has_role(role):
    return has_any_role(role)


def has_any_authority(*authorities):
    def decorator(func):
        def wrapper(*args, **kwargs):
            user = current_user._get_current_object()
            if user is None or not user.has_any_authority(*authorities):
                raise Unauthorized
            return func(*args, **kwargs)

        return update_wrapper(wrapper, func)

    return decorator


def has_authority(authority):
    return has_any_authority(authority)



# coding=utf-8

from functools import update_wrapper
import inspect
from typing import List

from flask import Blueprint as FlaskBlueprint, request, current_app, g
from guniflask.context import ApplicationEvent
from werkzeug.exceptions import BadRequest, InternalServerError
from werkzeug.routing import parse_rule

from guniflask.web.request_filter import RequestFilter, RequestFilterChain
from guniflask.beans.constructor_resolver import ConstructorResolver
from guniflask.web.filter_annotation import FilterChain
from guniflask.annotation.core import AnnotationUtils
from guniflask.beans.post_processor import BeanPostProcessorAdapter
from guniflask.web.bind_annotation import Blueprint, Route
from guniflask.utils.factory import instantiate_from_json, inspect_args
from guniflask.web.param_annotation import FieldInfo, RequestParam, PathVariable, \
    RequestParamInfo, PathVariableInfo, RequestBodyInfo, ContextParamInfo
from guniflask.web import param_annotation
from guniflask.beans.factory import BeanFactory, BeanFactoryAware
from guniflask.context.event_listener import ApplicationEventListener
from guniflask.context.event import ContextRefreshedEvent
from guniflask.web.filter_annotation import MethodFilter
from guniflask.utils.factory import resolve_arg_type, ArgType

__all__ = ['BlueprintPostProcessor']


class BlueprintPostProcessor(BeanPostProcessorAdapter, ApplicationEventListener, BeanFactoryAware):
    def __init__(self):
        self.blueprints = []
        self._filter_chain_resolver: FilterChainResolver = None
        self._method_filter_resolver = MethodFilterResolver()

    def set_bean_factory(self, bean_factory: BeanFactory):
        self._filter_chain_resolver = FilterChainResolver(bean_factory)

    def on_application_event(self, application_event: ApplicationEvent):
        if isinstance(application_event, ContextRefreshedEvent):
            self._register_blueprints()

    def post_process_after_initialization(self, bean, bean_name: str):
        bean_type = bean.__class__
        annotation = AnnotationUtils.get_annotation(bean_type, Blueprint)
        if annotation is not None:
            options = annotation['options'] or {}
            b = FlaskBlueprint(bean_name, bean.__module__,
                               url_prefix=annotation['url_prefix'], **options)
            annotation['blueprint'] = b
            for m in dir(bean):
                method = getattr(bean, m)
                a = AnnotationUtils.get_annotation(method, Route)
                if a is not None:
                    rule_options = a['options'] or {}
                    b.add_url_rule(a['rule'],
                                   endpoint=method.__name__,
                                   view_func=self.wrap_view_func(a['rule'], method),
                                   **rule_options)
                self._method_filter_resolver.resolve(b, method)
            self.blueprints.append(b)
            self._filter_chain_resolver.add_blueprint(bean_type)
        return bean

    def wrap_view_func(self, rule: str, method):
        params, param_names = self._resolve_method_parameters(rule, method)

        def wrapper(**kwargs):
            method_kwargs = self._resolve_method_kwargs(kwargs, params, param_names)
            return method(**method_kwargs)

        return update_wrapper(wrapper, method)

    def _register_blueprints(self):
        self._filter_chain_resolver.build()

        for b in self.blueprints:
            current_app.register_blueprint(b)

    def _resolve_method_parameters(self, rule: str, method):
        params = {}
        param_names = {}

        # resolve path variable in route
        path_variable_type = {}
        for converter, _, name in parse_rule(rule):
            if converter is None:
                continue
            if converter == 'int':
                path_variable_type[name] = int
            elif converter == 'float':
                path_variable_type[name] = float
            else:
                path_variable_type[name] = str

        args, type_hints = inspect_args(method)

        for arg, default in args.items():
            if default is inspect._empty:
                required = True
                default = None
            else:
                required = False
            if inspect.isfunction(default) and getattr(param_annotation, default.__name__, None) is default:
                default = default()
            if not isinstance(default, FieldInfo):
                if arg in path_variable_type:
                    annotation = PathVariable(dtype=None if arg in type_hints else path_variable_type[arg])
                else:
                    annotation = RequestParam()
                annotation.default = default
                default = annotation
            if arg in type_hints and default.dtype is None:
                default.dtype = type_hints[arg]
            if default.required is None:
                default.required = required
            params[arg] = default
            if default.name is not None:
                param_names[default.name] = arg

        return params, param_names

    def _resolve_method_kwargs(self, kwargs: dict, params: dict, param_names: dict) -> dict:
        result = {}
        for k, v in kwargs.items():
            if k in param_names:
                k = param_names[k]
            if k not in params:
                raise InternalServerError('Unhandled parameter: {}={}'.format(k, v))
            if params[k].dtype is not None:
                try:
                    v = params[k].dtype(v)
                except ValueError:
                    pass
            result[k] = v
        for k, p in params.items():
            if k in result:
                continue
            name = p.name or k
            if isinstance(p, PathVariableInfo):
                raise InternalServerError('No such path variable: {}'.format(name))

            if isinstance(p, RequestParamInfo):
                if name in request.args:
                    argc, etype = resolve_arg_type(p.dtype)
                    if argc is ArgType.LIST:
                        v = request.args.getlist(name)
                        if etype:
                            for i in range(len(v)):
                                v[i] = self._read_value(v[i], etype)
                    elif argc is ArgType.SINGLE:
                        v = request.args[name]
                        if p.dtype is not None:
                            v = self._read_value(v, p.dtype)
                    else:
                        raise ValueError('Unsupported request param type: {}'.format(p.dtype))

                    if v is not None:
                        result[k] = v
            elif isinstance(p, ContextParamInfo):
                name = p.name or k
                if name in g:
                    result[k] = g.get(name)
            elif isinstance(p, RequestBodyInfo):
                # FIXME: handle files, multi-parts, etc.
                data = request.json
                v = instantiate_from_json(data, dtype=p.dtype)
                if v is not None:
                    result[k] = v

            if k not in result:
                if p.required:
                    if isinstance(p, RequestBodyInfo):
                        raise BadRequest('Request body not given or in wrong format')
                    else:
                        raise BadRequest('Parameter not given: {}'.format(name))
                result[k] = p.default
        return result

    def _read_value(self, v: str, dtype: type):
        if v is None:
            return
        if dtype == bool:
            try:
                return bool(int(v))
            except (ValueError, TypeError):
                if v in ("True", "true"):
                    return True
                if v in ("False", "false"):
                    return False
        else:
            try:
                return dtype(v)
            except (ValueError, TypeError):
                pass


class FilterChainResolver:
    def __init__(self, bean_factory: BeanFactory):
        self._constructor_resolver = ConstructorResolver(bean_factory)
        self._blueprints = []

    def add_blueprint(self, bean_type):
        self._blueprints.append(bean_type)

    def build(self):
        for bean_type in self._blueprints:
            filter_chain_annotation = AnnotationUtils.get_annotation(bean_type, FilterChain)
            if filter_chain_annotation is not None:
                request_filters = self._resolve_request_filters(filter_chain_annotation['values'])
                if request_filters:
                    chain = RequestFilterChain()
                    for f in request_filters:
                        chain.add_request_filter(f)

                    blueprint_annotation = AnnotationUtils.get_annotation(bean_type, Blueprint)
                    b = blueprint_annotation['blueprint']
                    b.before_request(chain.before_request)
                    b.after_request(chain.after_request)

    def _resolve_request_filters(self, values) -> List[RequestFilter]:
        if not hasattr(values, '__iter__'):
            values = [values]
        result = []
        for v in values:
            if isinstance(v, RequestFilter):
                result.append(v)
            else:
                f = self._constructor_resolver.instantiate(v)
                assert isinstance(f, RequestFilter), 'Required a request filter, ' \
                                                     'got: {}'.format(type(RequestFilter).__name__)
        return result


class MethodFilterResolver:

    def resolve(self, blueprint, method):
        a = AnnotationUtils.get_annotation(method, MethodFilter)
        if a is None:
            return
        for v in a['values']:
            name = v['name']
            args = v['args']
            if args is None:
                getattr(blueprint, name)(method)
            else:
                getattr(blueprint, name)(*args)(method)



# coding=utf-8

import logging

from flask import Flask


def redirect_logger(name, logger):
    log = logging.getLogger(name)
    log.handlers = logger.handlers
    log.setLevel(logger.level)


def redirect_app_logger(app: Flask, logger):
    app.logger.handlers = logger.handlers
    app.logger.setLevel(logger.level)



# coding=utf-8

from abc import ABCMeta, abstractmethod

from guniflask.security_config.security_configurer import SecurityConfigurerAdapter
from guniflask.security.user_details_service import UserDetailsService

__all__ = ['UserDetailsAwareConfigurer']


class UserDetailsAwareConfigurer(SecurityConfigurerAdapter, metaclass=ABCMeta):
    @property
    @abstractmethod
    def user_details_service(self) -> UserDetailsService:
        pass



# coding=utf-8

from guniflask.annotation.core import AnnotationUtils
from guniflask.context.annotation import Include
from guniflask.security_config.web_security_config import WebSecurityConfiguration
from guniflask.security_config.authentication_config import AuthenticationConfiguration

__all__ = ['enable_authentication',
           'enable_web_security']


def enable_authentication(func):
    values = [AuthenticationConfiguration]
    AnnotationUtils.add_annotation(func, Include(values))
    return func


def enable_web_security(func):
    values = [WebSecurityConfiguration,
              AuthenticationConfiguration]
    AnnotationUtils.add_annotation(func, Include(values))
    return func



# coding=utf-8

import os
import tempfile
from os.path import join, exists
import fcntl

from guniflask.annotation.core import AnnotationMetadata
from guniflask.context.annotation import conditional
from guniflask.context.condition import Condition, ConditionContext
from guniflask.utils.process import get_master_pid

__all__ = ['global_singleton']


class GlobalSingletonCondition(Condition):
    global_singleton_conditions = {}

    def matches(self, context: ConditionContext, metadata: AnnotationMetadata) -> bool:
        singleton_id = self.generate_singleton_id(metadata)
        if singleton_id in self.global_singleton_conditions:
            return True

        temp_dir = join(tempfile.gettempdir(), 'guniflask', 'global_singleton_lock')
        if not exists(temp_dir):
            os.makedirs(temp_dir, exist_ok=True)

        fd = open(join(temp_dir, singleton_id), 'w')
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            return False
        self.global_singleton_conditions[singleton_id] = fd
        return True

    def generate_singleton_id(self, metadata: AnnotationMetadata) -> str:
        obj = metadata.source
        if hasattr(obj, '__module__'):
            module_name = obj.__module__
        else:
            module_name = None
        if hasattr(obj, '__name__'):
            obj_name = obj.__name__
        else:
            obj_name = None
        master_pid = get_master_pid()
        return '{}.{}_{}'.format(module_name, obj_name, master_pid)


def global_singleton(func):
    return conditional(GlobalSingletonCondition)(func)



# coding=utf-8

from flask import g

from guniflask.web.request_filter import RequestFilter
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.oauth2.token_extractor import TokenExtractor, BearerTokenExtractor
from guniflask.security.context import SecurityContext
from guniflask.oauth2.authentication_details import OAuth2AuthenticationDetails
from guniflask.security.authentication_token import AuthenticationToken

__all__ = ['OAuth2AuthenticationFilter']


class OAuth2AuthenticationFilter(RequestFilter):
    def __init__(self):
        self.authentication_manager: AuthenticationManager = None
        self.token_extractor: TokenExtractor = BearerTokenExtractor()

    def before_request(self):
        authentication = self.token_extractor.extract()
        if authentication:
            g.setdefault(OAuth2AuthenticationDetails.ACCESS_TOKEN_VALUE, authentication.principal)
            if isinstance(authentication, AuthenticationToken):
                authentication.details = OAuth2AuthenticationDetails()
            auth_result = self.authentication_manager.authenticate(authentication)
            SecurityContext.set_authentication(auth_result)



# coding=utf-8

from typing import Any

__all__ = [
    'FieldInfo',
    'RequestParam', 'RequestParamInfo',
    'PathVariable', 'PathVariableInfo',
    'RequestBody', 'RequestBodyInfo',
    'ContextParam', 'ContextParamInfo'
]


class FieldInfo:
    __slots__ = (
        'name',
        'default',
        'dtype',
        'description',
        'required',
        'extra',
    )

    def __init__(self, **kwargs):
        self.name = kwargs.pop('name', None)
        self.default = kwargs.pop('default', None)
        self.dtype = kwargs.pop('dtype', None)
        self.description = kwargs.pop('description', None)
        self.required = kwargs.pop('required', None)
        self.extra = kwargs


def RequestParam(name: str = None,
                 default: Any = None,
                 dtype: type = None,
                 description: str = None,
                 required: bool = None,
                 **extract: Any) -> Any:
    return RequestParamInfo(name=name,
                            default=default,
                            dtype=dtype,
                            description=description,
                            required=required,
                            **extract)


class RequestParamInfo(FieldInfo):
    pass


def PathVariable(name: str = None,
                 default: Any = None,
                 dtype: type = None,
                 description: str = None,
                 required: bool = None,
                 **extract: Any) -> Any:
    return PathVariableInfo(name=name,
                            default=default,
                            dtype=dtype,
                            description=description,
                            required=required,
                            **extract)


class PathVariableInfo(FieldInfo):
    pass


def RequestBody(name: str = None,
                default: Any = None,
                dtype: type = None,
                description: str = None,
                required: bool = None,
                **extract: Any) -> Any:
    return RequestBodyInfo(name=name,
                           default=default,
                           dtype=dtype,
                           description=description,
                           required=required,
                           **extract)


class RequestBodyInfo(FieldInfo):
    pass


def ContextParam(name: str = None,
                 default: Any = None,
                 dtype: type = None,
                 description: str = None,
                 required: bool = None,
                 **extract: Any) -> Any:
    return RequestParamInfo(name=name,
                            default=default,
                            dtype=dtype,
                            description=description,
                            required=required,
                            **extract)


class ContextParamInfo(FieldInfo):
    pass



# coding=utf-8

from guniflask.oauth2.token_granter import AbstractTokenGranter
from guniflask.oauth2.token_service import AuthorizationServerTokenServices
from guniflask.oauth2.client_details_service import ClientDetailsService
from guniflask.oauth2.request_factory import OAuth2RequestFactory

__all__ = ['ClientCredentialsTokenGranter']


class ClientCredentialsTokenGranter(AbstractTokenGranter):
    GRANT_TYPE = 'client_credentials'

    def __init__(self, token_services: AuthorizationServerTokenServices, client_details_service: ClientDetailsService,
                 request_factory: OAuth2RequestFactory):
        super().__init__(token_services, client_details_service, request_factory, self.GRANT_TYPE)
        self.allow_refresh = False

    def grant(self, grant_type, token_request):
        token = super().grant(grant_type, token_request)
        if token and not self.allow_refresh:
            token = token.copy()
            token.refresh_token = None
        return token



# coding=utf-8

from guniflask.oauth2.client_details import ClientDetails
from guniflask.oauth2.oauth2_utils import OAuth2Utils

__all__ = ['AuthorizationRequest', 'TokenRequest', 'OAuth2Request']


class BaseRequest:
    def __init__(self, client_id: str = None, scope: set = None, request_parameters: dict = None):
        self.client_id = client_id
        self.scope = set()
        if scope is not None:
            self.scope.update(scope)
        self.request_parameters = {}
        if request_parameters is not None:
            self.request_parameters.update(request_parameters)


class AuthorizationRequest(BaseRequest):
    def __init__(self, approval_parameters: dict = None, authorities: set = None, approved: bool = None,
                 resource_ids: set = None, redirect_uri: str = None, response_types: set = None, state: str = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.approval_parameters = {}
        if approval_parameters is not None:
            self.approval_parameters.update(approval_parameters)
        self.resource_ids = set()
        self.authorities = set()
        if authorities is not None:
            self.authorities.update(authorities)
        self.approved = False
        if approved is not None:
            self.approved = approved
        self.resource_ids = {}
        if resource_ids is not None:
            self.resource_ids = {}
        if resource_ids is not None:
            self.resource_ids.update(resource_ids)
        self.redirect_uri = redirect_uri
        self.response_types = set()
        if response_types is not None:
            self.response_types.update(response_types)
        self.state = state

    def set_from_client_details(self, client_details: ClientDetails):
        self.resource_ids = client_details.resource_ids
        self.authorities = client_details.authorities


class TokenRequest(BaseRequest):
    def __init__(self, grant_type: str = None, **kwargs):
        self.grant_type = grant_type
        super().__init__(**kwargs)


class OAuth2Request(BaseRequest):
    def __init__(self, authorities: set = None, approved: bool = None, resource_ids: set = None,
                 redirect_uri: str = None, response_types: set = None, **kwargs):
        super().__init__(**kwargs)
        self.authorities = set()
        if authorities is not None:
            self.authorities.update(authorities)
        self.approved = False
        if approved is not None:
            self.approved = approved
        self.resource_ids = {}
        if resource_ids is not None:
            self.resource_ids.update(resource_ids)
        self.redirect_uri = redirect_uri
        self.response_types = set()
        if response_types is not None:
            self.response_types.update(response_types)
        self.refresh: TokenRequest = None

    @property
    def grant_type(self):
        if OAuth2Utils.GRANT_TYPE in self.request_parameters:
            return self.request_parameters[OAuth2Utils.GRANT_TYPE]
        if OAuth2Utils.RESPONSE_TYPE in self.request_parameters:
            response = self.request_parameters[OAuth2Utils.RESPONSE_TYPE]
            if 'token' in response:
                return 'implicit'

    def copy(self):
        request = self.__class__(request_parameters=self.request_parameters,
                                 client_id=self.client_id,
                                 scope=self.scope,
                                 authorities=self.authorities,
                                 approved=self.approved,
                                 resource_ids=self.resource_ids,
                                 redirect_uri=self.redirect_uri,
                                 response_types=self.response_types)
        request.refresh = self.refresh
        return request

    def do_refresh(self, token_request: TokenRequest):
        request = self.copy()
        request.refresh = token_request
        return request

    def narrow_scope(self, scope: set):
        request = self.copy()
        request.scope = scope
        return request



# coding=utf-8

import re
from typing import Optional, Dict

from flask import request

__all__ = ['OAuth2Utils']


class OAuth2Utils:
    CLIENT_ID = 'client_id'
    STATE = 'state'
    SCOPE = 'scope'
    REDIRECT_URI = 'redirect_uri'
    RESPONSE_TYPE = 'response_type'
    USER_OAUTH_APPROVAL = 'user_oauth_approval'
    SCOPE_PREFIX = 'scope.'
    GRANT_TYPE = 'grant_type'

    @staticmethod
    def parse_parameter_list(values: str) -> set:
        result = set()
        if values is not None:
            values = values.strip()
            if values:
                s = re.split(r'\s+', values)
                result.update(s)
        return result

    @staticmethod
    def format_parameter_list(value: set) -> Optional[str]:
        if not value:
            return None
        return ' '.join(value)

    @staticmethod
    def get_request_parameters() -> Dict[str, str]:
        result = {}
        for k in request.args:
            result[k] = request.args.get(k)
        return result



# coding=utf-8

from abc import ABCMeta, abstractmethod

from apscheduler.schedulers.background import BackgroundScheduler
from guniflask.beans.factory_hook import DisposableBean, SmartInitializingSingleton

__all__ = ['AsyncExecutor', 'DefaultAsyncExecutor']


class AsyncExecutor(metaclass=ABCMeta):

    @abstractmethod
    def submit(self, task):
        pass


class DefaultAsyncExecutor(AsyncExecutor, SmartInitializingSingleton, DisposableBean):
    def __init__(self, scheduler=None):
        self._scheduler = scheduler or BackgroundScheduler()

    @property
    def scheduler(self):
        return self._scheduler

    def submit(self, task):
        self._scheduler.add_job(task)

    def start(self):
        self._scheduler.start(paused=False)

    def shutdown(self, wait=True):
        self._scheduler.shutdown(wait=wait)

    def after_singletons_instantiated(self):
        self.start()

    def destroy(self):
        self.shutdown(wait=False)



# coding=utf-8

from .abstract_endpoint import *
from .authentication import *
from .authentication_details import *
from .authentication_filter import *
from .authentication_manager import *
from .client_details import *
from .client_details_service import *
from .client_details_user_details_service import *
from .client_grant import *
from .errors import *
from .oauth2_utils import *
from .password_grant import *
from .refresh_grant import *
from .request import *
from .request_factory import *
from .token import *
from .token_converter import *
from .token_endpoint import *
from .token_extractor import *
from .token_granter import *
from .token_service import *
from .token_store import *

__all__ = (abstract_endpoint.__all__ +
           authentication.__all__ +
           authentication_details.__all__ +
           authentication_filter.__all__ +
           authentication_manager.__all__ +
           client_details.__all__ +
           client_details_service.__all__ +
           client_details_user_details_service.__all__ +
           client_grant.__all__ +
           errors.__all__ +
           oauth2_utils.__all__ +
           password_grant.__all__ +
           refresh_grant.__all__ +
           request.__all__ +
           request_factory.__all__ +
           token.__all__ +
           token_converter.__all__ +
           token_endpoint.__all__ +
           token_extractor.__all__ +
           token_granter.__all__ +
           token_service.__all__ +
           token_store.__all__)



# coding=utf-8

from typing import Optional

from flask import _request_ctx_stack

from guniflask.security.authentication import Authentication

__all__ = ['SecurityContext']


class SecurityContext:
    AUTHENTICATION = '__authentication'

    @classmethod
    def get_authentication(cls) -> Optional[Authentication]:
        ctx = _request_ctx_stack.top
        if ctx is not None:
            if not hasattr(ctx, cls.AUTHENTICATION):
                return None
            return getattr(ctx, cls.AUTHENTICATION)

    @classmethod
    def set_authentication(cls, authentication: Authentication):
        ctx = _request_ctx_stack.top
        if ctx is not None:
            setattr(ctx, cls.AUTHENTICATION, authentication)
